import torch
import time
import torchax
import torchax.interop
import os
import importlib
import sys
import logging
import sys

root = logging.getLogger()
root.setLevel(logging.DEBUG)

handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.DEBUG)
formatter = logging.Formatter(
    '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
handler.setFormatter(formatter)
root.addHandler(handler)

# NOTE: replace this patch below with your installation
TORCH_BENCH_PATH = os.path.expanduser('~/git/qihqi/benchmark')
# If your directory looks like this_file.py, benchmark/
sys.path.append(TORCH_BENCH_PATH)
model_name = "torchbenchmark.models.BERT_pytorch"  # replace this by the name of the model you're working on
module = importlib.import_module(model_name)
benchmark_cls = getattr(module, "Model", None)
benchmark = benchmark_cls(
    test="eval", device="cpu")  # test = train or eval device = cuda or cpu

model, example = benchmark.get_module()

env = torchax.default_env()
env.config.debug_print_each_op = False
model = env.to_xla(model)
example = env.to_xla(example)
with env:
  start = time.perf_counter()
  print(model(*example))
  end = time.perf_counter()
  print('Eager mode time', end - start)


def func_call(state, example):
  return torch.func.functional_call(model, state, example, tie_weights=False)


jitted = torchax.interop.jax_jit(func_call)
start = time.perf_counter()
print(func_call(model.state_dict(), example))
end = time.perf_counter()
print('Jitted mode time', end - start)



import torch
from torchax import tensor  # pylint: disable=unused-import
import torchax
import torchax.export

from .. import base_test_util
from . import llama_model
from . import model_exportable
from torch.utils import _pytree as pytree


class LlamaTest(base_test_util.TestCase):

  def test_can_run(self):
    with torchax.default_env():
      sample_args = (
          torch.randint(0, 32000, (1, 2048), device='jax:0'),
          torch.arange(0, 2048, device='jax:0'),
      )

      model_args = llama_model.ModelArgs(
          block_size=2048,
          vocab_size=32000,
          n_layer=2,
          n_head=4,
          dim=256,
      )
      m = llama_model.Transformer(model_args)
      m.to(torch.bfloat16)
      m.setup_caches(1, 2048)
      m = m.to('jax')

      print(m(*sample_args))

  def test_can_run_exportable(self):
    model_args = model_exportable.ModelArgs(
        vocab_size=32000,
        n_layers=2,
        n_heads=4,
        dim=256,
    )
    m = model_exportable.Transformer(model_args)
    context_length = 2048
    input_shape_prefill = (1, context_length)
    input_shape_decode = (1, 1)

    def make_cache(args, batch_size):
      n_kv_heads = args.n_heads if args.n_kv_heads is None else args.n_kv_heads
      n_local_heads = args.n_heads
      n_local_kv_heads = n_kv_heads
      n_rep = n_local_heads // n_local_kv_heads
      head_dim = args.dim // args.n_heads
      res = []
      for i in range(args.n_layers):
        if batch_size is None:
          size = (
              args.max_seq_len,
              n_local_kv_heads,
              head_dim,
          )
        else:
          size = (
              batch_size,
              args.max_seq_len,
              n_local_kv_heads,
              head_dim,
          )
        res.append(
            (torch.zeros(
                size,
                dtype=torch.bfloat16 if args.bf16_enable else torch.float),
             torch.zeros(
                 size,
                 dtype=torch.bfloat16 if args.bf16_enable else torch.float)))
      return res

    prefill_caches = make_cache(model_args, 1)

    sample_input_prefill = (
        torch.randint(0, 1000, input_shape_prefill,
                      dtype=torch.int32),  # len seq length
        torch.arange(0, context_length, dtype=torch.int32),  # input indexes
        torch.arange(0, context_length, dtype=torch.int32),  # context indexes
        prefill_caches,
        True,  # prefil
    )
    with torch.no_grad():
      m_prefill = torch.export.export(m, sample_input_prefill)

    weights, mj_prefill = torchax.export.exported_program_to_jax(m_prefill)
    env = torchax.default_env()
    sample_inputs = env.t2j_copy(sample_input_prefill)
    print('Prefill', mj_prefill(weights, sample_inputs))

    sample_input_decode = (
        torch.randint(0, 1000, input_shape_decode,
                      dtype=torch.int32),  # len = 1
        torch.tensor([0], dtype=torch.int32),
        torch.roll(torch.arange(context_length, dtype=torch.int32), 1, 0),
        prefill_caches,
        False  # prefill
    )
    with torch.no_grad():
      m_decode = torch.export.export(m, sample_input_decode)
    weights, mj_decode = torchax.export.exported_program_to_jax(m_decode)
    sample_inputs = env.t2j_copy(sample_input_decode)
    print('Decode', mj_decode(weights, sample_inputs))


if __name__ == "__main__":
  base_test_util.main()



import functools

import torch
from time import time
from diffusers import DiffusionPipeline
from torch.utils import _pytree as pytree

import torchax
import torchax.functions
from torchax.extra import torch_view, jax_view

import jax
import torch.func


class CompiledModule:

  def __init__(self, model):
    weights = model.state_dict()
    weights.update(model.named_parameters())
    self._weights = pytree.tree_map_only(torch.Tensor,
                                         torchax.tensor.move_to_device, weights)
    self._model = model

    self._func_jitted_torch = None  #torch_view(func_mod_jitted)

  def _maybe_move_tensor(self, tensor):
    if isinstance(
        tensor, torch.Tensor) and not isinstance(tensor, torchax.tensor.Tensor):
      return torchax.tensor.move_to_device(tensor)
    return tensor

  def _make_jitted(self, args, kwargs):
    static = []
    for i, a in enumerate(args):
      if not isinstance(a, torch.Tensor):
        static.append(i + 1)  # weight is 0
    static_argnames = []
    for k, v in kwargs.items():
      if not isinstance(v, torch.Tensor):
        static_argnames.append(k)

    def f(weights, *args, **kwargs):
      weights, args, kwargs = torchax.tensor.wrap((weights, args, kwargs))
      with torchax.functions.XLAFunctionMode(), torchax.tensor.XLADispatchMode(
      ):
        res = torch.func.functional_call(self._model, weights, args, kwargs)
        if isinstance(res, tuple) and len(res) == 1:
          res = res[0]
      return torchax.tensor.unwrap(res)

    fjit = jax.jit(f, static_argnames=tuple(static_argnames))
    return torch_view(fjit)

  def forward(self, *args, **kwargs):
    (args, kwargs) = pytree.tree_map(self._maybe_move_tensor, (args, kwargs))
    if self._func_jitted_torch is None:
      self._func_jitted_torch = self._make_jitted(args, kwargs)
    return self._func_jitted_torch(self._weights, *args, **kwargs)

  def __call__(self, *args, **kwargs):
    return self.forward(*args, **kwargs)

  def __getattr__(self, key):
    return getattr(self._model, key)


def compile_pipe(pipe):
  pipe.text_encoder = CompiledModule(pipe.text_encoder)
  pipe.text_encoder_2 = CompiledModule(pipe.text_encoder_2)
  pipe.unet = CompiledModule(pipe.unet)
  pipe.vae = CompiledModule(pipe.vae)


def main():
  pipe = DiffusionPipeline.from_pretrained(
      # "stabilityai/stable-diffusion-xl-base-0.9",
      "stabilityai/stable-diffusion-xl-base-1.0",
      use_safetensors=True,
  )
  compile_pipe(pipe)

  global_bs = 10
  inference_steps = 20
  resol = 1024
  prompts = ["a photo of an astronaut riding a horse on mars"] * global_bs
  print(
      f'global batch size {global_bs}',
      f'inference steps {inference_steps}',
      f'Image resolution {resol}',
      flush=True)

  iters = 5
  for i in range(iters):
    prompt = prompts
    # print('per device prompts len',len(prompt))
    # prompt = prompts[rank]
    start = time()
    image = pipe(
        prompt, num_inference_steps=inference_steps, height=resol,
        width=resol).images[0]
    print(f'Step {i} inference time {time()-start} sec', flush=True)


if __name__ == '__main__':
  main()



import unittest
import torch
import torch.nn.functional as F
from torch.library import Library, impl, impl_abstract
import torchax
import torchax.export
from torchax.ops import jaten
from torchax.ops import jlibrary
# Create a `mylib` library which has a basic SDPA op.
m = Library("mylib", "DEF")
m.define("scaled_dot_product_attention(Tensor q, Tensor k, Tensor v) -> Tensor")


@impl(m, "scaled_dot_product_attention", "CompositeExplicitAutograd")
def _mylib_scaled_dot_product_attention(q, k, v):
  """Basic scaled dot product attention without all the flags/features."""
  q = q.transpose(1, 2)
  k = k.transpose(1, 2)
  v = v.transpose(1, 2)
  y = F.scaled_dot_product_attention(
      q,
      k,
      v,
      dropout_p=0,
      is_causal=False,
      scale=None,
  )
  return y.transpose(1, 2)


@impl_abstract("mylib::scaled_dot_product_attention")
def _mylib_scaled_dot_product_attention_meta(q, k, v):
  return torch.empty_like(q)


# Register library op as a composite for export using the `@impl` method
# for a torch decomposition.
jlibrary.register_torch_composite(
    "mylib.scaled_dot_product_attention", _mylib_scaled_dot_product_attention,
    torch.ops.mylib.scaled_dot_product_attention,
    torch.ops.mylib.scaled_dot_product_attention.default)

# Also register ATen softmax as a composite for export in the `mylib` library
# using the JAX ATen decomposition from `jaten`.
jlibrary.register_jax_composite(
    "mylib.softmax",
    jaten._aten_softmax,
    torch.ops.aten._softmax,
    static_argnums=1  # Required by JAX jit
)


class LibraryTest(unittest.TestCase):

  def setUp(self):
    torch.manual_seed(0)
    torchax.default_env().config.use_torch_native_for_cpu_tensor = False

  def test_basic_sdpa_library(self):

    class CustomOpExample(torch.nn.Module):

      def forward(self, q, k, v):
        x = torch.ops.mylib.scaled_dot_product_attention(q, k, v)
        x = x + 1
        return x

    # Export and check for composite operations
    model = CustomOpExample()
    arg = torch.rand(32, 8, 128, 64)
    args = (
        arg,
        arg,
        arg,
    )

    exported = torch.export.export(model, args=args)
    weights, stablehlo = torchax.export.exported_program_to_stablehlo(exported)
    module_str = str(stablehlo.mlir_module())

    ## TODO Update this machinery from producing function calls to producing
    ## stablehlo.composite ops.
    self.assertIn("call @mylib.scaled_dot_product_attention", module_str)
    self.assertIn("call @mylib.softmax", module_str)


if __name__ == '__main__':
  unittest.main()



import dataclasses
import logging
from torchax.types import JaxCallable, TorchCallable

from typing import Union, Dict


@dataclasses.dataclass
class Operator:
  torch_op: TorchCallable
  func: Union[TorchCallable, JaxCallable]
  is_jax_function: bool
  is_user_defined: bool
  needs_env: bool
  is_view_op: bool


all_aten_ops: Dict[TorchCallable, Operator] = {}
all_torch_functions: Dict[TorchCallable, Operator] = {}


def register_torch_dispatch_op(aten_op,
                               impl_callable,
                               is_jax_function=True,
                               is_user_defined=False,
                               needs_env=False,
                               is_view_op=False):
  op = Operator(
      aten_op,
      impl_callable,
      is_jax_function=is_jax_function,
      is_user_defined=is_user_defined,
      needs_env=needs_env,
      is_view_op=is_view_op)
  if aten_op in all_aten_ops:
    logging.warning(f'Duplicate op registration for {aten_op}')
  all_aten_ops[aten_op] = op
  return impl_callable


def register_torch_function_op(torch_func,
                               impl_callable,
                               is_jax_function=True,
                               is_user_defined=False,
                               needs_env=False,
                               is_view_op=False):
  op = Operator(
      torch_func,
      impl_callable,
      is_jax_function=is_jax_function,
      is_user_defined=is_user_defined,
      needs_env=needs_env,
      is_view_op=is_view_op)
  all_torch_functions[torch_func] = op
  return impl_callable



def _is_in_bad_fork():
  return False


def manual_seed_all(seed):
  pass


def device_count():
  return 1


def get_rng_state():
  return []


def set_rng_state(new_state, device):
  pass


def is_available():
  return True


def current_device():
  return 0



import os, torch, torch.nn as nn, torch.utils.data as data, torchvision as tv
import lightning as L

encoder = nn.Sequential(nn.Linear(28 * 28, 128), nn.ReLU(), nn.Linear(128, 3))
decoder = nn.Sequential(nn.Linear(3, 128), nn.ReLU(), nn.Linear(128, 28 * 28))


class LitAutoEncoder(L.LightningModule):

  def __init__(self, encoder, decoder):
    super().__init__()
    self.encoder, self.decoder = encoder, decoder

  def training_step(self, batch, batch_idx):
    x, y = batch
    x = x.view(x.size(0), -1)
    z = self.encoder(x)
    x_hat = self.decoder(z)
    loss = nn.functional.mse_loss(x_hat, x)
    self.log("train_loss", loss)
    return loss

  def configure_optimizers(self):
    return torch.optim.Adam(self.parameters(), lr=1e-3)


dataset = tv.datasets.MNIST(
    ".", download=True, transform=tv.transforms.ToTensor())

# Lightning will automatically use all available GPUs!
trainer = L.Trainer()
# trainer.fit(LitAutoEncoder(encoder, decoder), data.DataLoader(dataset, batch_size=64))

# ==== above is the lightning example from
# https://lightning.ai/pytorch-lightning

import torchax
from torchax.interop import jax_view, torch_view
import jax
import optax


class JaxTrainer:

  def __init__(self):
    pass

  def torch_opt_to_jax_opt(self, torch_opt):
    # TODO: Can convert optimizer instead of using a jax one
    return optax.adam(0.001)

  def fit(self, lightning_mod, data_loader):

    xla_env = torchax.default_env()

    def lightning_mod_loss(weights: jax.Array, data: jax.Array, batch_id):
      """returns loss"""
      weights, data = torch_view((weights, data))
      lightning_mod.load_state_dict(weights, assign=True)
      with xla_env:
        loss = lightning_mod.training_step(data, batch_id)
      return jax_view(loss)

    jax_weights = jax_view(xla_env.to_xla(lightning_mod.state_dict()))
    jax_optimizer = self.torch_opt_to_jax_opt(
        lightning_mod.configure_optimizers())
    opt_state = jax_optimizer.init(jax_weights)
    grad_fn = jax.jit(jax.value_and_grad(lightning_mod_loss))

    for bid in range(3):
      for item in data_loader:
        xla_data = jax_view(xla_env.to_xla(item))
        loss, grads = grad_fn(jax_weights, xla_data, bid)
        updates, opt_state = jax_optimizer.update(grads, opt_state)
        jax_weights = optax.apply_updates(jax_weights, updates)
        print('current_loss', loss)


print('-----------------')
trainer_jax = JaxTrainer()
trainer_jax.fit(
    LitAutoEncoder(encoder, decoder), data.DataLoader(dataset, batch_size=64))



import jax.numpy as jnp
import jax
from jax.experimental.pallas.ops.tpu import flash_attention

import torchax
from jax.experimental import mesh_utils
from torchax.ops.jtorch import _tpu_flash_attention

env = torchax.default_env()
jax.config.update('jax_enable_x64', False)
env._mesh = jax.sharding.Mesh(
    mesh_utils.create_device_mesh((4,)),
    axis_names=("fsdp",),
)
env.use_flash_attention = True

from torch.nn import functional as F


def attn(q, k, v):
  q, k, v = env.j2t_iso((q, k, v))
  with env:
    x = F.scaled_dot_product_attention(q, k, v, is_causal=True)
  x = env.t2j_iso(x)
  return jnp.sum(x)


import torch


class M(torch.nn.Module):

  def __init__(self):
    super().__init__()
    self.a = torch.nn.Linear(10, 10)

  def forward(self, x):
    return self.a(x)


m = M()
from torchax.interop import JittableModule

mjit = JittableModule(m)

from torch.nn.utils import stateless


def f(weights, x):
  res = mjit.functional_call('forward', weights, {}, (x,))
  return torch.sum(res)


def crossent(x, y):
  x, y = env.j2t_iso((x, y))
  res = torch.func.functional_call(m, x, (y,))
  return env.t2j_iso(res)


graded = jax.value_and_grad(attn)

shape = (4, 32, 128, 32)
q = jnp.ones(shape, dtype='bfloat16')
v = jnp.ones(shape, dtype='bfloat16')
k = jnp.ones(shape, dtype='bfloat16')

env = torchax.default_env()
weights = env.t2j_iso(env.to_xla(mjit.params))

from torchax.interop import jax_view

#print(jax.jit(graded).lower(q, v, k).as_text())
print(
    jax.jit(jax.grad(jax_view(f))).lower(weights,
                                         jax.ShapeDtypeStruct(
                                             (10,), 'float32')).as_text())



import torch
import torchax

env = torchax.default_env()
env.config.debug_print_each_op = True
env.config.debug_accuracy_for_each_op = True

with env:
  y = torch.tensor([1, 5, 10])
  print(torch.trapezoid(y))
  print(torch.trapz(y, y))



from jax import dlpack as jaxdl
import jax.numpy as jnp
import numpy
import torch
import torch.func
import torch.utils.dlpack as torchdl
import torch.utils._mode_utils as mode_utils


def t2j(t, use_dlpack=True):
  is_bool = False
  if t.dtype == torch.bool:
    is_bool = True
    t = t.to(torch.int8)

  t = t.to_dense()

  if not t.is_contiguous():
    t = t.contiguous()

  res = None
  if use_dlpack:
    try:
      res = jaxdl.from_dlpack(t)
    except Exception:
      pass

  if res is None:
    # https://github.com/google/jax/issues/7657
    # https://github.com/google/jax/issues/17784
    if t.dtype == torch.bfloat16:
      nparray = (t.cpu().detach().to(torch.float32).numpy()
                )  # numpy don't support bfloat16
    else:
      nparray = t.cpu().detach().numpy()
    res = jnp.asarray(nparray)
    if t.dtype == torch.bfloat16:
      res = res.astype(jnp.bfloat16)

  if is_bool:
    res = res.astype(jnp.bool_)
  return res


def j2t(x, use_dlpack=True):
  with mode_utils.no_dispatch(), torch._C.DisableTorchFunction():
    res = None
    if use_dlpack:
      try:
        dl = jaxdl.to_dlpack(x)
        res = torchdl.from_dlpack(dl)
      except Exception:
        res = None

    orig_dtype = None
    if res is None:
      orig_dtype = None
      if x.dtype == jnp.bfloat16.dtype:
        orig_dtype = x.dtype
        x = x.astype(jnp.float32.dtype)
      res = torch.from_numpy(numpy.asarray(x))

    if x.dtype == jnp.bool_:
      res = res.to(torch.bool)

    if orig_dtype is not None:
      res = res.to(j2t_dtype(orig_dtype))
    return res


TORCH_DTYPE_TO_JAX = {
    # NO_MAPPING        : jnp.float0.dtype (signless scalar int),
    torch.bool:
        jnp.bool_.dtype,
    # NO_MAPPING        : jnp.int4.dtype,
    torch.int8:
        jnp.int8.dtype,
    torch.int16:
        jnp.int16.dtype,
    torch.int32:
        jnp.int32.dtype,
    torch.int64:
        jnp.int64.dtype,
    torch.long:
        jnp.int64.dtype,
    # NO_MAPPING        : jnp.uint4
    torch.uint8:
        jnp.uint8.dtype,
    torch.uint16:
        jnp.uint16.dtype,
    torch.uint32:
        jnp.uint32.dtype,
    torch.uint64:
        jnp.uint64.dtype,
    # NO_MAPPING        : jnp.float8_e4m3b11fnuz.dtype,
    torch.float8_e4m3fn:
        jnp.float8_e4m3fn.dtype,
    # NO_MAPPING        : jnp.float8_e4m3fnuz.dtype,
    torch.float8_e5m2:
        jnp.float8_e5m2.dtype,
    # NO_MAPPING        : jnp.float8_e5m2fnuz.dtype,
    torch.bfloat16:
        jnp.bfloat16.dtype,
    torch.half:
        jnp.float16.dtype,
    torch.float16:
        jnp.float16.dtype,
    torch.float32:
        jnp.float32.dtype,
    torch.float64:
        jnp.float64.dtype,
    torch.double:
        jnp.double.dtype,
    torch.complex64:
        jnp.complex64.dtype,
    torch.complex128:
        jnp.complex128.dtype,
    None:
        None,
}

JAX_DTYPE_TO_TORCH = {value: key for key, value in TORCH_DTYPE_TO_JAX.items()}
# Add imprecise mappings for some JAX dtypes which don't have torch analogues
JAX_DTYPE_TO_TORCH[jnp.dtype('int4')] = torch.int8
JAX_DTYPE_TO_TORCH[jnp.dtype('uint4')] = torch.uint8


def t2j_dtype(dtype):
  if dtype not in TORCH_DTYPE_TO_JAX:
    raise RuntimeError(
        f'Attempting to convert unknown type: {dtype} to jax type,')
  return TORCH_DTYPE_TO_JAX[dtype]


def j2t_dtype(dtype):
  if dtype not in JAX_DTYPE_TO_TORCH:
    raise RuntimeError(
        f'Attempting to convert unknown type: {dtype} to torch type,')
  return JAX_DTYPE_TO_TORCH[dtype]



import jax
import jax.numpy as jnp
import torch
from torch.nn.functional import linear

## Goal 1. Illustrate that class attr for torch modules
## dont play nice with jax.jit
## How we solve it.


class Linear(torch.nn.Module):

  def __init__(self):
    super().__init__()
    self.weight = torch.nn.Parameter(torch.randn(1000, 1000))
    self.bias = torch.nn.Parameter(torch.randn(1000))

  def forward(self, X):
    return linear(X, self.weight, self.bias)


# Running with torch native

print('---- example 1 -----')
m = Linear()
x = torch.randn(2, 1000)
print(m(x))
print(m.forward(x))

with torch.inference_mode():
  # with torch.no_grad():
  print(m.forward(x))

print('---- example 2 -----')

import torchax

env = torchax.default_env()

with env:
  m2 = Linear()
  x = torch.randn(2, 1000)
  print(m2.forward(x))

print('---- example 3 -----')
# where is the jax jit?

# m2 is a callable that takes in Tensor and returns Tensor
#    m2: (Tensor -> Tensor)

# suppose t2j (Tensor -> jax.Array) "unwraps the XLATensor"
# suppose j2t (jax.Array  -> Tensor) "wraps the XLATensor"
from torchax import tensor
import jax


def t2j(torch_tensor: tensor.Tensor) -> jax.Array:
  return torch_tensor._elem


def j2t(jax_array: jax.Array) -> tensor.Tensor:
  return tensor.Tensor(jax_array, env)


# # further notice t2j(j2t(x)) == x; j2t(t2j(x)) == x


def jax_m(X: jax.Array):
  X_jax = j2t(X)
  res = m2(X_jax)
  return t2j(res)


jax_x = jnp.ones((10, 1000))
print(jax_m(jax_x))

## Let f: Tensor -> Tensor
## There is a function g: jax.Array -> jax.Array;
##   g = x |-> j2t (f (t2j(x))). OR,
##   g = j2t . f . t2j (. denotes function composition)
# The correspondence f -> g is an isomorphism too.

jitted_jax_m = jax.jit(jax_m)
print(jitted_jax_m(jax_x))
print(jitted_jax_m.lower(jax_x).as_text())

from torch.utils import _pytree as pytree


def jax_m_functional(states, X):
  states_torch = pytree.tree_map(j2t, states)
  X = j2t(X)
  old_state_dict = m2.state_dict()
  m2.load_state_dict(states_torch, assign=True, strict=False)
  res = m2(X)
  m2.load_state_dict(old_state_dict, assign=True, strict=False)
  return t2j(res)


jax_weights = {
    'weight': m2.weight._elem,
    'bias': m2.bias._elem,
}

jitted_jax_m_functional = jax.jit(jax_m_functional)
print(jitted_jax_m_functional.lower(jax_weights, jax_x).as_text())

# ## interop module

# print('---- exmaple 4 ----')
# import torchax.interop

# def m_functional(states, x):
#   return torch.func.functional_call(m2, states, x)

# with jax.checking_leaks():
#   print(torchax.interop.jax_jit(m_functional)(m2.state_dict(), x))

# # Experiment if time:
# # 1. torch buffer persistence = False
# # 2. torch attr



import unittest
import torchax
import torch
from torch.testing._internal.common_utils import TestCase


class TestMutations(TestCase):

  def setUp(self):
    self.env = torchax.tensor.Environment()

  def test_add(self):
    with self.env:
      x = torch.tensor([1, 2, 3], dtype=torch.int32)
      y = torch.tensor([4, 5, 6], dtype=torch.int32)
      x.add_(y)
      self.assertEqual(x, torch.tensor([5, 7, 9], dtype=torch.int32))

  def test_sub(self):
    with self.env:
      x = torch.tensor([1, 2, 3], dtype=torch.int32)
      y = torch.tensor([4, 5, 6], dtype=torch.int32)
      x.sub_(y)
      self.assertEqual(x, torch.tensor([-3, -3, -3], dtype=torch.int32))

  def test_mul(self):
    with self.env:
      x = torch.tensor([1, 2, 3], dtype=torch.int32)
      y = torch.tensor([4, 5, 6], dtype=torch.int32)

      x.mul_(y)
      self.assertEqual(x, torch.tensor([4, 10, 18], dtype=torch.int32))


if __name__ == '__main__':
  unittest.main()



# pylint: disable
"""Utilities for exporting a torch program to jax/stablehlo."""
import copy
from typing import Any, Dict, Tuple
import torch
from torch.utils import _pytree as pytree
import torchax
from torchax import tensor
from torchax.ops import ops_registry, mappings
from torchax import decompositions
import jax
import jax.export
import sympy

DEBUG = False


class JaxInterpreter(torch.fx.Interpreter):
  """Experimental."""

  def __init__(self, graph_module):
    super().__init__(graph_module)
    import torchax.ops.jaten
    import torchax.ops.jtorch

  def call_function(self, target, args: Tuple, kwargs: Dict) -> Any:
    if not isinstance(target,
                      (torch._ops.OpOverloadPacket, torch._ops.OpOverload)):
      return super().call_function(target, args, kwargs)

    if DEBUG:
      print('Running ', target.name(), '--------')

    op = ops_registry.all_aten_ops.get(target)
    if op is None:
      op = ops_registry.all_aten_ops.get(target.overloadpacket)
    assert op is not None, target
    assert op.is_jax_function, op
    if op is None:
      op = ops_registry.all_aten_ops.get(target.overloadpacket)
    if op is None:
      print(target.name(), target.tags)
      raise RuntimeError('No lowering found for', target.name())
    return op.func(*args, **kwargs)

  def run_node(self, n) -> Any:
    res = super().run_node(n)
    if DEBUG:
      if n.op == 'call_function':
        if hasattr(res, 'shape'):
          print('Meta:', n.meta.get('val').shape, 'REAL: ', res.shape)
    return res


from torch._decomp import get_decompositions
import torch._refs

_extra_decomp = get_decompositions([torch.ops.aten.unfold])


def _extract_states_from_exported_program(exported_model):
  # NOTE call convention: (parameters, buffers, user_inputs)
  param_and_buffer_keys = exported_model.graph_signature.parameters + exported_model.graph_signature.buffers
  state_dict = copy.copy(exported_model.state_dict)
  if (constants := getattr(exported_model, 'constants', None)) is not None:
    state_dict.update(constants)
  param_buffer_values = list(state_dict[key] for key in param_and_buffer_keys)

  if hasattr(exported_model.graph_signature, "lifted_tensor_constants"):
    for name in exported_model.graph_signature.lifted_tensor_constants:
      param_buffer_values.append(exported_model.tensor_constants[name])

  return param_and_buffer_keys, param_buffer_values


def exported_program_to_jax(exported_program, export_raw: bool = False):
  """returns a pytree of jax arrays(state), and

  a callable(func) that is jax function.

  func(state, input) would be how you call it.
  """
  if torch.__version__ >= '2.2':
    # torch version 2.1 didn't expose this yet
    exported_program = exported_program.run_decompositions()
    exported_program = exported_program.run_decompositions(
        decompositions.DECOMPOSITIONS)
  if DEBUG:
    print(exported_program.graph_module.code)

  names, states = _extract_states_from_exported_program(exported_program)

  def _extract_args(args, kwargs):
    flat_args, received_spec = pytree.tree_flatten(
        (args, kwargs))  # type: ignore[possibly-undefined]
    return flat_args

  num_mutations = len(exported_program.graph_signature.buffers_to_mutate)

  def func(states, inputs):
    args = _extract_args(inputs, {})
    res = JaxInterpreter(exported_program.graph_module).run(
        *states,
        *args,
        enable_io_processing=False,
    )
    res = res[num_mutations:]
    return res

  if export_raw:
    return names, states, func
  env = torchax.default_env()
  states = env.t2j_copy(states)
  return states, func


def extract_avals(exported):
  """Return JAX Abstract Value shapes for all input parameters of the exported
  program. This supports dynamic batch dimensions, including with constraints.
  """

  def _to_aval(arg_meta, symbolic_shapes):
    """Convet from torch type to jax abstract value for export tracing
    """

    def _get_dim(d):
      if isinstance(d, torch.SymInt):
        return symbolic_shapes[str(d)]
      return d

    val = arg_meta['val']
    is_scalar = isinstance(val, float) or isinstance(val, int) or isinstance(
        val, bool)
    if is_scalar:
      return jax.ShapeDtypeStruct([], type(arg_meta['val']))

    tensor_meta = arg_meta['tensor_meta']
    shape = [_get_dim(d) for d in tensor_meta.shape]
    return jax.ShapeDtypeStruct(shape, mappings.t2j_dtype(tensor_meta.dtype))

  def _get_inputs(exported):
    """Return placeholders with input metadata"""
    placeholders = [p for p in exported.graph.nodes if p.op == "placeholder"]
    input_placeholders = [
        p for p, s in zip(placeholders, exported.graph_signature.input_specs)
        if s.kind == torch.export.graph_signature.InputKind.USER_INPUT
    ]
    return input_placeholders

  def _build_symbolic_shapes(range_constraints):
    """Convert torch SymInt to JAX symbolic_shape and stores in a map using the
    string name of the torch symbolic int.

    TODO: There is probably a better way of storing a key for a symbolic int.
    This value needs to be looked up again in `_to_aval` to figure out which
    JAX symbolic to map to for a given torch tensor.
    """
    if len(range_constraints) == 0:
      return None

    def _build_symbolic_constraints(symbol_name, torch_constraint):
      """Convert torch SymInt constraints to string for JAX symbolic_shape
      Using sympy may be overkill here, currently PyTorch only uses ValueRanges
      which allow specifying the min and the max of a value, for example:
        torch.export.Dim("a", min=5, max=10)
         ==> ("a >= 5", "a <= 10",)
      """
      if not isinstance(torch_constraint, torch.utils._sympy.value_ranges.
                        ValueRanges) or torch_constraint.is_bool:
        raise TypeError(
            f"No symbolic constraint handler for: {torch_constraint}")

      constraints = []
      symbol = sympy.Symbol(symbol_name)
      if torch_constraint.lower != 2:
        constraints.append(symbol >= torch_constraint.lower)
      from sympy.core.singleton import S
      if not torch_constraint.upper.is_infinite and torch_constraint.upper is not S.IntInfinity:
        constraints.append(symbol <= torch_constraint.upper)

      return tuple(sympy.pretty(c, use_unicode=False) for c in constraints)

    def _build_symbolic_shape(sym, constraint, free_symbols):
      """Returns a JAX symbolic shape for a given symbol and constraint

      There are two possible sympy `sym` inputs:
        1. Symbol - (s0) These can have custom constraints.
        2. Expr - (s0*2) These apply the expr to s0's constraints, cannot override.

        Currently support is limited to operations with a symbol and and int,
        in `torch/export/dynamic_shapes.py`:
        "Only increasing linear operations with integer coefficients are supported."
      """
      symbol_name = str(sym)
      constraints = _build_symbolic_constraints(symbol_name, constraint)
      if sym.is_symbol:
        symbolic_shape = jax.export.symbolic_shape(
            symbol_name, constraints=constraints)
      else:
        assert len(sym.free_symbols) > 0
        scope = free_symbols[str(list(sym.free_symbols)[0])].scope
        symbolic_shape = jax.export.symbolic_shape(symbol_name, scope=scope)
      assert len(symbolic_shape) == 1
      return symbolic_shape[0]

    # Populate symbol variables before expressions, exprs need to use the same
    # Symbolic scope as the variable they operate on. Expressions can only be
    # integer compuations on symbol variables, so each symbol variable is OK to
    # have its own scope.
    symbolic_shapes = {}
    symbol_variables = [
        (s, v) for s, v in range_constraints.items() if s.is_symbol
    ]
    symbol_exprs = [
        (s, v) for s, v in range_constraints.items() if not s.is_symbol
    ]
    for sym, constraint in symbol_variables + symbol_exprs:
      symbolic_shape = _build_symbolic_shape(sym, constraint, symbolic_shapes)
      symbolic_shapes[str(sym)] = symbolic_shape
    return symbolic_shapes

  symbolic_shapes = _build_symbolic_shapes(exported.range_constraints)
  args = _get_inputs(exported)

  if DEBUG:
    print('Inputs to aval:', args, '--------')
    print('Symbolic shapes:', symbolic_shapes)
    for arg in args:
      print('Meta2Aval', arg.meta, '--> ', _to_aval(arg.meta, symbolic_shapes))

  return [_to_aval(arg.meta, symbolic_shapes) for arg in args]


def exported_program_to_stablehlo(exported_program):
  """Replacement for torch_xla.stablehlo.exported_program_to_stablehlo

  Convert a program exported via torch.export to StableHLO.

  This supports dynamic dimension sizes and generates explicit checks for
  dynamo guards in the IR using shape_assertion custom_call ops.
  """
  weights, func = exported_program_to_jax(exported_program)
  jax_avals = extract_avals(exported_program)
  jax_export = jax.export.export(jax.jit(func))(weights, (jax_avals,))
  return weights, jax_export



# From: https://github.com/google/gemma_pytorch/blob/main/gemma/tokenizer.py

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from typing import List, Optional

from sentencepiece import SentencePieceProcessor


class Tokenizer:

  def __init__(self, model_path: Optional[str]):
    # Reload tokenizer.
    assert os.path.isfile(model_path), model_path
    self.sp_model = SentencePieceProcessor(model_file=model_path)

    # BOS / EOS token IDs.
    self.n_words: int = self.sp_model.vocab_size()
    self.bos_id: int = self.sp_model.bos_id()
    self.eos_id: int = self.sp_model.eos_id()
    self.pad_id: int = self.sp_model.pad_id()
    assert self.sp_model.vocab_size() == self.sp_model.get_piece_size()

  def encode(self, s: str, bos: bool = True, eos: bool = False) -> List[int]:
    """Converts a string into a list of tokens."""
    assert isinstance(s, str)
    t = self.sp_model.encode(s)
    if bos:
      t = [self.bos_id] + t
    if eos:
      t = t + [self.eos_id]
    return t

  def decode(self, t: List[int]) -> str:
    """Converts a list of tokens into a string."""
    return self.sp_model.decode(t)



import functools
import jax
import jax.numpy as jnp
import numpy as np
import torch
from torchax.ops import mappings
from torchax.view import View
from torchax import types
import sys

from typing import Callable, Optional, ParamSpec, Concatenate


class InplaceOp:

  def __init__(self,
               functional_op,
               replace=False,
               position_to_mutate=0,
               is_jax_func=False):
    self.functional = functional_op
    self.replace = replace
    self.position_to_mutate = position_to_mutate
    self.is_jax_func = is_jax_func

  def __call__(self, *args, **kwargs):
    to_mutate = args[self.position_to_mutate]
    view_value = to_mutate
    if isinstance(to_mutate, View):
      view_value = to_mutate.torch()
      # Convert the target View to a Tensor, and
      # leave the rest args as is. If other args are
      # also View, they will be converted to tensors
      # in the self.functional dispatch.
    env = view_value._env
    if self.is_jax_func:
      view_value, args, kwargs = env.t2j_iso((view_value, args, kwargs))
      new_value_jax = self.functional(view_value, *args[1:], **kwargs)
      new_value = env.j2t_iso(new_value_jax)
    else:
      new_value = self.functional(view_value, *args[1:], **kwargs)

    if isinstance(to_mutate, View):
      to_mutate.update(new_value)
    else:
      if self.replace:
        to_mutate._elem = new_value._elem
      else:
        to_mutate.copy_(new_value)
    return to_mutate


class OutVariant:

  def __call__(self, *args, **kwargs):
    to_mutate = kwargs['out']
    del kwargs['out']
    to_mutate._elem = self.functional(*args, **kwargs)._elem
    return to_mutate


P = ParamSpec('P')


def convert_dtype(use_default_dtype: bool = True):
  """Converts `dtype` kwarg of function from torch to JAX.

  Args:
    use_default_dtype: Whether to use torch default dtype if none is provided.

  Returns:
    A decorator that wraps a JAX implementation of a torch function.
  """

  def decorator(func: types.TorchCallable):

    @functools.wraps(func)
    def wrapper(*args: P.args,
                dtype: Optional[torch.dtype] = None,
                **kwargs: P.kwargs):
      if not dtype and use_default_dtype:
        dtype = torch.get_default_dtype()
      if isinstance(dtype, torch.dtype):
        jax_dtype = mappings.t2j_dtype(dtype)
      else:
        jax_dtype = dtype

      return func(*args, dtype=jax_dtype, **kwargs)

    return wrapper

  return decorator


def maybe_convert_constant_dtype(val: Optional[types.JaxValue],
                                 dtype: Optional[jnp.dtype]):
  """Optionally converts scalar constant's dtype using `numpy`

  Use in cases where you require a constant and can't handle a traced array.
  """
  if val and dtype:
    if isinstance(val, jax.Array):
      return maybe_convert_constant_dtype(val.item(), dtype)

    return np.array(val, dtype)

  return val


def promote_int_input(f: Callable[Concatenate[jax.Array, P], types.JaxValue]):
  """If the first argument is an int array, promote it to float32."""

  @functools.wraps(f)
  def wrapper(x: jax.Array, *args: P.args, **kwargs: P.kwargs):
    if x.dtype in [jnp.int8, jnp.int16, jnp.int32, jnp.int64]:
      x = x.astype(mappings.t2j_dtype(torch.get_default_dtype()))

    return f(x, *args, **kwargs)

  return wrapper


def foreach_loop(seq: jax.Array,
                 fn: Callable[[jax.Array, jax.Array], jax.Array],
                 init_val=0.0):
  """Run `fn` for each element of 1D array `seq`.

  Similar to `functools.reduce`, but implemented with `jax.lax.fori_loop`."""
  assert len(seq.shape) == 1
  return jax.lax.fori_loop(0, len(seq), lambda i, carry: fn(carry, seq[i]),
                           init_val)



import os
import time
import logging
from typing import Tuple
from collections import defaultdict
import functools
import torch
import torch.nn.functional
from torch.utils import _pytree as pytree
import splash_attn
import helper

import torchax as tx
import torchax.interop
import torchax.train
from torchax.interop import jax_view, torch_view, JittableModule
import jax
import jax.numpy as jnp
from jax.experimental import shard_map
from jax.experimental import mesh_utils
from jax.sharding import NamedSharding
import optax

from torchtitan.models.llama import llama3_configs
from torchtitan.models.llama import model as titan

P = jax.sharding.PartitionSpec

num_global_devices = jax.device_count()
num_local_devices = jax.local_device_count()


def sharded_device_put(tensor: jax.Array, sharding) -> jax.Array:
  if isinstance(tensor, tuple):
    return tuple(sharded_device_put(t, sharding) for t in tensor)

  if num_global_devices == num_local_devices:
    return jax.device_put(tensor, sharding)

  # NOTE: at here, num_global_devices != num_local_devices
  # meaning we are in multi-host setup. Each host will run the same process
  # and each process only need to handle the devices accessible to this host.
  shape = tensor.shape
  x_split = [
      jax.device_put(tensor[i], device)
      for device, i in sharding.addressable_devices_indices_map(shape).items()
  ]
  return jax.make_array_from_single_device_arrays(shape, sharding, x_split)


sharding_map_original = {
    "freqs_cis": (),  #  torch.complex64 (2048, 64)
    "tok_embeddings.weight":
        ('fsdp', 'tp'),  #  torch.float32 (vocab_size, 4096)
    "layers.*.attention.wo.weight": ('fsdp', 'tp'),  #  torch.int8 (4096, 4096)
    "layers.*.attention.wq.weight": ('tp', 'fsdp'),  #  torch.int8 (4096, 4096)
    "layers.*.attention.wk.weight": ('tp', 'fsdp'),  #  torch.int8 (4096, 4096)
    "layers.*.attention.wv.weight": ('tp', 'fsdp'),  #  torch.int8 (4096, 4096)
    "layers.*.feed_forward.w1.weight":
        ('tp', 'fsdp'),  #  torch.float32 (11008, 4096)
    "layers.*.feed_forward.w2.weight":
        ('fsdp', 'tp'),  #  torch.float32 (4096, 11008)
    "layers.*.feed_forward.w3.weight":
        ('tp', 'fsdp'),  #  torch.float32 (11008, 4096)
    "layers.*.attention_norm.weight": ('fsdp',),  #  torch.float32 (4096,)
    "layers.*.ffn_norm.weight": ('fsdp',),  #  torch.float32 (4096,)
    "norm.weight": ('fsdp',),  #  torch.float32 (4096,)
    "output.weight": ('tp', 'fsdp'),  #  torch.float32 (vocab_size, 4096)
}

sharding_map_scan = {
    "freqs_cis": (),  #  torch.complex64 (2048, 64)
    # ParallelEmbedding for llama2; VocabParallelEmbedding for 3
    "tok_embeddings.weight":
        ('tp', 'fsdp'),  #  torch.float32 (vocab_size, 4096)
    "layers.params.attention___wo___weight":
        (None, 'fsdp', 'tp'),  #  torch.int8 (n, 4096, 4096)
    "layers.params.attention___wq___weight":
        (None, 'tp', 'fsdp'),  #  torch.int8 (n, 4096, 4096)
    "layers.params.attention___wk___weight":
        (None, 'tp', 'fsdp'),  #  torch.int8 (n, 4096, 4096)
    "layers.params.attention___wv___weight":
        (None, 'tp', 'fsdp'),  #  torch.int8 (n, 4096, 4096)
    "layers.params.feed_forward___w1___weight":
        (None, 'tp', 'fsdp'),  #  torch.float32 (n, 11008, 4096)
    "layers.params.feed_forward___w2___weight":
        (None, 'fsdp', 'tp'),  #  torch.float32 (n, 4096, 11008)
    "layers.params.feed_forward___w3___weight":
        (None, 'tp', 'fsdp'),  #  torch.float32 (n, 11008, 4096)
    "layers.params.attention_norm___weight": (
        None,
        'fsdp',
    ),  #  torch.float32 (n, 4096,)
    "layers.params.ffn_norm___weight": (
        None,
        'fsdp',
    ),  #  torch.float32 (n, 4096,)
    "norm.weight": ('fsdp',),  #  torch.float32 (4096,)
    "output.weight": ('tp', 'fsdp'),  #  torch.float32 (vocab_size, 4096)
}

sharding_map_scan_fsdp = {
    "freqs_cis": (),  #  torch.complex64 (2048, 64)
    # ParallelEmbedding for llama2; VocabParallelEmbedding for 3
    "tok_embeddings.weight": ('fsdp',),  #  torch.float32 (vocab_size, 4096)
    "layers.params.attention___wo___weight":
        (None, 'fsdp'),  #  torch.int8 (n, 4096, 4096)
    "layers.params.attention___wq___weight":
        (None, 'fsdp'),  #  torch.int8 (n, 4096, 4096)
    "layers.params.attention___wk___weight":
        (None, 'fsdp'),  #  torch.int8 (n, 4096, 4096)
    "layers.params.attention___wv___weight":
        (None, 'fsdp'),  #  torch.int8 (n, 4096, 4096)
    "layers.params.feed_forward___w1___weight":
        (None, 'fsdp'),  #  torch.float32 (n, 11008, 4096)
    "layers.params.feed_forward___w2___weight":
        (None, 'fsdp'),  #  torch.float32 (n, 4096, 11008)
    "layers.params.feed_forward___w3___weight":
        (None, 'fsdp'),  #  torch.float32 (n, 11008, 4096)
    "layers.params.attention_norm___weight": (
        None,
        'fsdp',
    ),  #  torch.float32 (n, 4096,)
    "layers.params.ffn_norm___weight": (
        None,
        'fsdp',
    ),  #  torch.float32 (n, 4096,)
    "norm.weight": ('fsdp',),  #  torch.float32 (4096,)
    "output.weight": ('fsdp',),  #  torch.float32 (vocab_size, 4096)
}


class Trainer:

  def __init__(self, mesh):
    self.mesh = mesh
    self.x_sharding = jax.sharding.NamedSharding(self.mesh, P('fsdp'))
    self.replicated = jax.sharding.NamedSharding(self.mesh, P())

  def fit(self, model, loss_fn, data_loader):
    xla_env = torchax.default_env()
    jax.config.update('jax_enable_x64', False)
    xla_env._mesh = self.mesh
    xla_env.use_flash_attention = True

    jittable_mod = JittableModule(model)

    # split the params to the n devices

    # model_fn is responsible to shard if needed
    # to do FSDP one shards the first input args and output
    # on the batch dimension
    def model_fn(weights, buffers, args):
      return jittable_mod.functional_call('forward', weights, buffers, args)

    jax_optimizer = optax.sgd(0.01)
    opt_state = torch_view(jax_optimizer.init(jax_view(jittable_mod.params)))

    #opt_state = torchax.interop.call_jax(jax_optimizer.init, jittable_mod.params)

    train_step = torchax.train.make_train_step(
        model_fn,
        loss_fn,
        jax_optimizer,
        remat_policy=jax.checkpoint_policies.offload_dot_with_no_batch_dims(
            'device', 'pinned_host'))

    print('Begining training')
    s = time.perf_counter()
    jax.profiler.start_trace('/tmp/tensorboard')
    print('start training')
    min_loop_time = 10000
    for i, item in enumerate(data_loader):
      inputs, labels = item
      # Move them to jax device
      inputs = inputs.to('jax')
      labels = labels.to('jax')

      # Shard them on batch dim for fsdp
      inputs.apply_jax_(sharded_device_put, self.x_sharding)
      labels.apply_jax_(sharded_device_put, self.x_sharding)

      if i == 0:
        train_step = helper.compile_step_func(train_step, jittable_mod.params,
                                              jittable_mod.buffers, opt_state,
                                              inputs, labels, self.mesh)

      print('INPUT shape', inputs.shape)
      step_start = time.perf_counter()
      loss, jittable_mod.params, opt_state = train_step(jittable_mod.params,
                                                        jittable_mod.buffers,
                                                        opt_state, inputs,
                                                        labels)
      # wait for iteration to finish to measure time
      torchax.interop.call_jax(jax.block_until_ready,
                               (loss, jittable_mod.params))
      step_end = time.perf_counter()
      print(i, 'loss', loss, 'step latency: ', step_end - step_start)
      loop_time = step_end - step_start
      min_loop_time = min(min_loop_time, loop_time)
      print('======')
      if i >= 3:
        break
    jax.profiler.stop_trace()
    return min_loop_time


def _process_sharding_name(name):
  """Replace integers in param name with *.

  Presumably all layers should have the same sharding.
  """

  def is_integer(t):
    try:
      int(t)
      return True
    # pylint: disable-next=all
    except:  # noqa: E722
      return False

  tokens = name.split(".")
  for i, t in enumerate(tokens):
    if is_integer(t):
      tokens[i] = "*"
  return ".".join(tokens)


def create_sharded_weights(model, mesh, sharding_map):
  res = {}
  env = torchax.default_env()
  for name, weight_meta in model.state_dict().items():
    sharding_spec = sharding_map.get(_process_sharding_name(name))
    if sharding_spec is None:
      print('Skipping weight:', name)
      continue
    sharding = NamedSharding(mesh, P(*sharding_spec))
    with jax.default_device(jax.devices('cpu')[0]):
      weight_torch = torch.randn(weight_meta.shape, dtype=weight_meta.dtype)
      weight_jax = torchax.default_env().to_xla(weight_torch).jax()
    #print(name, weight.shape, weight.dtype)
    res[name] = env.j2t_iso(
        jax.make_array_from_callback(weight_jax.shape, sharding,
                                     lambda a: weight_jax[a]))
  return res


def fake_dataloader(size, seqlen, batch_size):
  for _ in range(size):
    x = torch.randint(0, 32000, (batch_size, seqlen), device='cpu')
    yield x, (x + 1) % 32000


def main(
    model_type='8B',
    batch_size=8,
    seqlen=2048,
    override_num_layers=-1,
    use_scan=True,
    tp_parallelism=1,
):
  torchax.enable_globally()
  torchax.enable_performance_mode()
  #logging.getLogger("jax").setLevel(logging.DEBUG)
  print(f"Running with parameters {locals()}")

  fsdp = num_global_devices // tp_parallelism
  mesh = jax.make_mesh((fsdp, tp_parallelism), ('fsdp', 'tp'))
  if use_scan:
    # using scan the individial weights will have shape (num_layers, w, h)
    sharding_map = sharding_map_scan_fsdp
  else:
    sharding_map = sharding_map_original

  env = torchax.default_env()
  env.config.use_tpu_flash_attention = True
  env.config.shmap_flash_attention = True
  env._mesh = mesh  # this is the mesh used by flash attention pallas kernel

  args = llama3_configs[model_type]
  # Note: torchtitan's upstream config did not specify this value
  args.vocab_size = 128256
  args.max_seq_len = seqlen
  if override_num_layers > 0:
    args.n_layers = override_num_layers

  # Note: because a single device don't have enough HBM memory
  # nor enough CPU memory to hold the parameters. We instantiate
  # the model on meta then manually initialize then shard each param
  torch.set_default_dtype(torch.bfloat16)
  with torch.device('meta'):
    gpt = titan.Transformer(args)

  with torch.device('cpu'):
    # need actual value for freqs_cis
    freqs_cis = gpt._precompute_freqs_cis()

  if use_scan:
    checkpoint_policy = jax.checkpoint_policies.offload_dot_with_no_batch_dims(
        'device', 'pinned_host')
    gpt = TransfomerWithScan(gpt, checkpoint_policy)

  state_dict = dict(gpt.state_dict())
  state_dict.pop('freqs_cis')  # dont shard freqs_cis
  state_dict = create_sharded_weights(gpt, mesh, sharding_map)
  replicated = jax.sharding.NamedSharding(mesh, P())

  state_dict['freqs_cis'] = freqs_cis.to('jax').apply_jax(
      jax.device_put, replicated)
  gpt.load_state_dict(state_dict, assign=True)

  train_loader = fake_dataloader(10, seqlen, batch_size)

  # NOTE: overriding attention to capture mesh and sharding info
  partition = P('fsdp', 'tp', None, None)
  attention = functools.partial(splash_attn.tpu_splash_attention, mesh,
                                partition, True)
  attention = jax.jit(attention)

  def custom_attention(query,
                       key,
                       value,
                       attn_mask=None,
                       dropout_p=0.0,
                       is_causal=False,
                       scale=None,
                       enable_gqa=False):
    #  batch, num of head, seq, dim
    jk, jq, jv = jax_view((query, key, value))
    res = attention(jk, jq, jv, None)
    return torch_view(res)

  env.override_op_definition(torch.nn.functional.scaled_dot_product_attention,
                             custom_attention)

  def loss_fn(logits, y):
    num_tokens = logits.shape[-1]
    logits = logits.reshape(-1, num_tokens)
    y = y.reshape(-1)
    return torch.nn.functional.cross_entropy(logits, y)

  with mesh:
    trainer = Trainer(mesh)
    return trainer.fit(gpt, loss_fn, train_loader)


class TransfomerWithScan(torch.nn.Module):

  def __init__(self, old_transformer, checkpoint_policy):
    super().__init__()
    self.tok_embeddings = old_transformer.tok_embeddings
    self.norm = old_transformer.norm
    self.output = old_transformer.output
    self.layers = torchax.train.ScannedModule(
        list(old_transformer.layers.values()), checkpoint_policy)

    self.register_buffer('freqs_cis', old_transformer.freqs_cis)

  def forward(self, tokens: torch.Tensor):
    """
        Perform a forward pass through the Transformer model.

        Args:
            tokens (torch.Tensor): Input token indices.

        Returns:
            torch.Tensor: Output logits after applying the Transformer model.

        """
    # passthrough for nonexistent layers, allows easy configuration of pipeline parallel stages
    h = self.tok_embeddings(tokens) if self.tok_embeddings else tokens

    # for layer in self.layers.values():
    #     h = layer(h, self.freqs_cis)

    h = self.layers(h, self.freqs_cis)

    h = self.norm(h) if self.norm else h
    output = self.output(h) if self.output else h
    return output


if __name__ == '__main__':
  import fire
  fire.Fire(main)



from typing import Callable, Any, Union, ParamSpec, TypeAlias
import torch
import jax
import jax.numpy as jnp
import sys

P = ParamSpec('P')

TorchValue: TypeAlias = Union[torch.Tensor, torch.dtype, 'TorchCallable', Any]
TorchCallable: TypeAlias = Callable[P, TorchValue]
JaxValue: TypeAlias = Union[jax.Array, jnp.dtype, 'JaxCallable', Any]
JaxCallable: TypeAlias = Callable[P, JaxValue]


import torch
from torch import nn
import torchax
from . import base_test_util


class CustomConv1(torch.nn.Module):

  def __init__(
      self,
      channels_conv1=3,
      width_conv1=3,
      channels_conv2=5,
      width_conv2=5,
      hidden_layer_size=50,
  ):
    super(CustomConv1, self).__init__()
    self.conv1 = nn.Conv1d(1, channels_conv1, width_conv1)
    self.conv2 = nn.Conv1d(channels_conv1, channels_conv2, width_conv2)
    self.fc1 = nn.Linear(hidden_layer_size, 2)

  def forward(self, x):
    x = nn.functional.max_pool1d(nn.functional.relu(self.conv1(x)), 2, stride=2)
    x = nn.functional.max_pool1d(nn.functional.relu(self.conv2(x)), 2, stride=2)
    x = torch.flatten(x, 1)
    x = nn.functional.softmax(self.fc1(x), dim=1)
    return x


class CustomConv2(nn.Module):

  def __init__(self):
    super().__init__()
    inp = 4
    out = 16

    self.conv = nn.Conv2d(inp, out, kernel_size=3, padding=1)

    # This is supposed to be a squeeze and excitation block.
    self.avg_pool = nn.AdaptiveAvgPool2d((1, 1))

    self.scale = nn.Sequential(nn.Linear(out, out), nn.Sigmoid())

  def forward(self, x):
    x = self.conv(x)

    b = x.shape[0]
    ap = self.avg_pool(x).view(b, -1)
    ap = self.scale(ap)
    ap = ap.view(b, -1, 1, 1)

    return x * ap


class ConvTest(base_test_util.TestCase):

  def test_conv1(self):
    env = torchax.default_env()
    m = CustomConv1()
    arg = torch.randn((20, 1, 50))
    res = m(arg)

    jax_weights, jax_func = torchax.extract_jax(m)
    arg = env.t2j_copy(arg)
    res2 = jax_func(jax_weights, (arg,))
    res2_torch = env.j2t_copy(res2)
    self.assertTrue(torch.allclose(res, res2_torch))

  def test_conv2(self):
    env = torchax.default_env()
    m = CustomConv2()
    arg = torch.randn((20, 4, 50, 100))
    res = m(arg)
    jax_weights, jax_func = torchax.extract_jax(m)
    arg = env.t2j_copy(arg)
    res2 = jax_func(jax_weights, (arg,))
    res2_torch = env.j2t_copy(res2)
    self.assertTrue(torch.allclose(res, res2_torch, atol=1e-4, rtol=1e-4))


if __name__ == '__main__':
  base_test_util.main()



import jax
import jax.numpy as jnp


def cubic_kernel(x, a=-0.75):
  """Cubic kernel with a = -0.75 (PyTorch-like Keys kernel)"""
  absx = jnp.abs(x)
  x2 = absx * absx
  x3 = x2 * absx
  cond1 = (absx <= 1)
  cond2 = (absx > 1) & (absx < 2)
  f1 = (a + 2) * x3 - (a + 3) * x2 + 1
  f2 = a * x3 - 5 * a * x2 + 8 * a * absx - 4 * a
  return jnp.where(cond1, f1, jnp.where(cond2, f2, 0.0))


def compute_contribs(in_size,
                     out_size,
                     scale,
                     support=2.0,
                     align_corners=False,
                     dtype=None):
  if align_corners:
    if out_size == 1:
      in_coords = jnp.zeros((1,), dtype=dtype)
    else:
      in_coords = jnp.linspace(0, in_size - 1, out_size, dtype=dtype)
  else:
    out_coords = jnp.arange(out_size, dtype=dtype) + 0.5
    in_coords = out_coords / scale - 0.5

  left_idx = jnp.floor(in_coords).astype(jnp.int32) - 1
  idxs = left_idx[:, None] + jnp.arange(4)

  dx = in_coords[:, None] - idxs

  weights = cubic_kernel(dx)

  weights = weights / jnp.sum(weights, axis=1, keepdims=True)
  return idxs, weights


def gather_weights(img, idxs, axis):
  """Safely gather with boundary handling"""
  idxs = jnp.clip(idxs, 0, img.shape[axis] - 1)
  return jnp.take(img, idxs, axis=axis)


def interpolate_along_axis_bchw(img, idxs, weights, axis):
  """
    Interpolate along H (axis=2) or W (axis=3) for tensor (B, C, H, W).
    idxs: (out_size, 4) int32 indices
    weights: (out_size, 4) float32 weights
    """
  assert axis in (2, 3), "Axis must be 2 (H) or 3 (W)"
  out_size = idxs.shape[0]
  k = idxs.shape[1]  # Typically 4 for cubic

  # Clip to input bounds
  idxs = jnp.clip(idxs, 0, img.shape[axis] - 1)  # (out_size, 4)

  def gather_and_weight(i):
    idx = idxs[i]  # (4,)
    w = weights[i]  # (4,)

    def gather_one(offset):
      return jnp.take(img, idx[offset], axis=axis)  # shape (B, C, H, W)

    gathered = jnp.stack([gather_one(o) for o in range(k)],
                         axis=0)  # (4, B, C, H, W)
    weighted = jnp.tensordot(w, gathered, axes=(0, 0))  # (B, C, H, W)
    return weighted

  out = jax.vmap(gather_and_weight)(
      jnp.arange(out_size))  # (out_size, B, C, H, W)

  # Move the interpolated axis back into place
  if axis == 2:  # interpolated over H
    return jnp.moveaxis(out, 0, 2)  # (B, C, out_H, W)
  else:  # axis == 3, interpolated over W
    return jnp.moveaxis(out, 0, 3)  # (B, C, H, out_W)


def interpolate_bicubic_no_aa(img, out_h, out_w, align_corners=False):
  h, w = img.shape[-2:]
  if align_corners and out_h > 1:
    scale_y = (h - 1) / (out_h - 1)
  else:
    scale_y = out_h / h

  if align_corners and out_w > 1:
    scale_x = (w - 1) / (out_w - 1)
  else:
    scale_x = out_w / w

  idxs_y, weights_y = compute_contribs(
      h,
      out_h,
      scale_y,
      align_corners=align_corners,
      dtype=img.dtype,
  )
  tmp = interpolate_along_axis_bchw(img, idxs_y, weights_y, axis=2)

  idxs_x, weights_x = compute_contribs(
      w,
      out_w,
      scale_x,
      align_corners=align_corners,
      dtype=img.dtype,
  )
  out = interpolate_along_axis_bchw(tmp, idxs_x, weights_x, axis=3)
  return out



import unittest
from torchax import interop
import torch


class MyAwesomeModel(torch.nn.Module):
  pass


class EvenMoreAwesomeModel(torch.nn.Module):
  pass


class JittableModuleTest(unittest.TestCase):

  def test_isinstance_works(self):

    # Export and check for composite operations
    model = MyAwesomeModel()
    jittable_module = interop.JittableModule(model)

    # jittable_module should remain an instance of MyAwesomeModel logicailly
    assert isinstance(jittable_module, MyAwesomeModel)

  def test_isinstance_does_not_mix(self):

    # Export and check for composite operations
    JittableAwesomeModel = interop.JittableModule(MyAwesomeModel())
    JittableMoreAwesomeModel = interop.JittableModule(EvenMoreAwesomeModel())

    # jittable_module should remain an instance of MyAwesomeModel logicailly
    assert isinstance(JittableAwesomeModel, MyAwesomeModel)
    assert not isinstance(JittableAwesomeModel, EvenMoreAwesomeModel)
    assert isinstance(JittableMoreAwesomeModel, EvenMoreAwesomeModel)
    assert not isinstance(JittableMoreAwesomeModel, MyAwesomeModel)


if __name__ == '__main__':
  unittest.main()



import torch
import unittest
import torchax
from torch.utils import _pytree as pytree
from . import config
from . import model as gemma


class GemmaTest(unittest.TestCase):

  def setup(self):
    torch.manual_seed(0)

  def test_gemma(self):
    mconfig = config.GemmaConfig(
        num_hidden_layers=3,
        num_attention_heads=8,
        num_key_value_heads=1,
        hidden_size=256,
        intermediate_size=16384,
        dtype=torch.float32)
    model = gemma.GemmaForCausalLM(mconfig)
    batch_size = 1
    max_seq_len = 1000
    min_prompt_len = 1000
    device = 'cpu'
    pad_id = -1
    temperature = 0.8
    top_k = 100
    top_p = 1.0

    # prepare inputs
    token_ids_tensor = torch.randint(
        0, max_seq_len, (batch_size, max_seq_len), dtype=torch.int64)

    # build KV caches
    kv_caches = []
    for _ in range(model.config.num_hidden_layers):
      size = (batch_size, max_seq_len, model.config.num_key_value_heads,
              model.config.head_dim)
      dtype = model.config.get_dtype()
      k_cache = torch.zeros(size=size, dtype=dtype, device=device)
      v_cache = torch.zeros(size=size, dtype=dtype, device=device)
      kv_caches.append((k_cache, v_cache))

    token_ids_tensor = token_ids_tensor.to(device)
    prompt_mask_tensor = torch.ones_like(token_ids_tensor)
    input_positions_tensor = torch.arange(
        0, min_prompt_len, dtype=torch.int64).to(device)
    mask_tensor = torch.full((1, 1, max_seq_len, max_seq_len),
                             -2.3819763e38).to(torch.float)
    mask_tensor = torch.triu(mask_tensor, diagonal=1).to(device)
    curr_mask_tensor = mask_tensor.index_select(2, input_positions_tensor)
    output_positions_tensor = torch.LongTensor([min_prompt_len - 1]).to(device)
    temperatures_tensor = torch.FloatTensor([temperature] *
                                            batch_size).to(device)
    top_ps_tensor = torch.FloatTensor([top_p] * batch_size).to(device)
    top_ks_tensor = torch.LongTensor([top_k] * batch_size).to(device)
    output_index = torch.tensor(min_prompt_len, dtype=torch.int64).to(device)

    inputs = (
        token_ids_tensor,
        input_positions_tensor,
        None,  # kv_write_indexes
        kv_caches,
        mask_tensor,
        output_positions_tensor,
        temperatures_tensor,
        top_ps_tensor,
        top_ks_tensor,
    )

    weights, jax_func = torchax.extract_jax(model)
    env = torchax.default_env()
    inputs_jax = env.t2j_copy(inputs)

    import jax
    print(jax.jit(jax_func)(weights, inputs_jax))


if __name__ == '__main__':
  unittest.main()



import torch
import torchax
import torchax.export
from . import base_test_util


class AddOne(torch.nn.Module):

  def __init__(self):
    super().__init__()

  def forward(self, a):
    return a + 1


class ConcatAddModel(torch.nn.Module):

  def __init__(self):
    super().__init__()

  def forward(self, a, b):
    a = torch.concat([a, a], dim=0)
    return a + b


class SymbolicShapeTest(base_test_util.TestCase):
  """Test possible symbolic shape computations that upstream torch export can
  emit. Seems to be currently limited to a few binary math operations where one
  operand is a symbolic variable/expr and the other is a constant integer.
  """

  def setUp(self):
    torch.manual_seed(0)

  def test_constraints_min_max(self):
    """Test a model with basic min/max dimension restrictions
    """

    # Arg shapes are a=s0{<=10}, b=s0*2
    model = AddOne()
    args = (torch.rand(5),)
    sym_a = torch.export.Dim("a", min=3, max=10)
    dynamic_shapes = ({0: sym_a},)

    with torch.no_grad():
      exported = torch.export.export(
          model, args=args, dynamic_shapes=dynamic_shapes)
    weights, stablehlo = torchax.export.exported_program_to_stablehlo(exported)
    module_str = str(stablehlo.mlir_module())

    self.assertRegex(module_str, r"stablehlo.constant.*3")
    self.assertRegex(module_str, r"shape_assertion.*s[0-9]+ >= 3")
    self.assertRegex(module_str, r"stablehlo.constant.*10")
    self.assertRegex(module_str, r"shape_assertion.*s[0-9]+ <= 10")

  def test_constraints_multiply(self):
    """Test a model with a slightly more complex constraint, where the input
    shapes are determined by an equation of the other, in this case input shapes
    are s0{<=10} and s0*2.
    """
    # Arg shapes are a=s0{<=10}, b=s0*2
    model = ConcatAddModel()
    args = (torch.rand(2), torch.rand(4))
    sym_a = torch.export.Dim("a", max=10)
    sym_b = sym_a * 2
    dynamic_shapes = ({0: sym_a}, {0: sym_b})

    with torch.no_grad():
      exported = torch.export.export(
          model, args=args, dynamic_shapes=dynamic_shapes)
    weights, stablehlo = torchax.export.exported_program_to_stablehlo(exported)
    module_str = str(stablehlo.mlir_module())

    self.assertRegex(module_str, r"stablehlo.constant.*10")
    self.assertRegex(module_str, r"shape_assertion.*s[0-9]+ <= 10")
    self.assertRegex(module_str, r"stablehlo.constant.*2")
    self.assertRegex(module_str, r"shape_assertion.*2\*s[0-9]+")

  def test_constraint_indirection(self):
    """Test a model where none of the shapes are directly symbolic variables
    but all are expressions of symints that don't appear directly in the model.
    """

    # Arg shapes are b=s0{<=10}*2
    args = (torch.randn(10, 10),)
    model = AddOne()
    sym_a = torch.export.Dim("a", max=10)
    sym_b = sym_a * 2
    dynamic_shapes = ({0: sym_b},)

    with torch.no_grad():
      exported = torch.export.export(
          model, args=args, dynamic_shapes=dynamic_shapes)
    weights, stablehlo = torchax.export.exported_program_to_stablehlo(exported)
    module_str = str(stablehlo.mlir_module())

    self.assertRegex(module_str, r"shape_assertion.*s[0-9]+ <= 10")
    self.assertRegex(module_str, r"shape_assertion.*2\*s[0-9]+")


if __name__ == "__main__":
  base_test_util.main()



import dataclasses


@dataclasses.dataclass
class Configuration:
  debug_print_each_op: bool = False
  debug_accuracy_for_each_op: bool = False
  debug_mixed_tensor: bool = False
  debug_print_each_op_operands: bool = False

  use_int32_for_index: bool = False

  # If true, we will convert Views into torchax.Tensors eagerly
  force_materialize_views: bool = False

  # Use DLPack for converting jax.Arrays <-> and torch.Tensor
  use_dlpack_for_data_conversion: bool = False

  # Flash attention
  use_tpu_flash_attention: bool = False
  shmap_flash_attention: bool = False

  # device
  treat_cuda_as_jax_device: bool = True
  use_torch_native_for_cpu_tensor: bool = True
  internal_respect_torch_return_dtypes: bool = False



import os
import tempfile
import numpy as np
import tensorflow as tf
import torch
import torch.nn.functional as F
import torchax

from torchax import tf_integration
from . import base_test_util


class Interpolate(torch.nn.Module):

  def forward(self, masks: torch.Tensor) -> torch.Tensor:
    masks = F.interpolate(
        masks,
        size=(500, 500),
        mode="bilinear",
        align_corners=False,
    )
    return masks


class TfIntegrationTest(base_test_util.TestCase):

  def setUp(self):
    torch.manual_seed(0)
    torchax.enable_accuracy_mode()

  def test_interpolate(self):
    """Simple model roundtripped through TF savedmodel"""

    # Create model
    arg = (torch.randn(3, 3, 200, 200),)
    pt_model = Interpolate()

    # Export to SavedModel
    with tempfile.TemporaryDirectory() as tempdir:
      sm_path = os.path.join(tempdir, "interpolate.savedmodel")
      tf_integration.save_torch_module_as_tf_saved_model(pt_model, arg, sm_path)

      # Reload SM and compare results with PT results
      loaded_model = tf.saved_model.load(sm_path)
      pt_res = pt_model(*arg)
      tf_res = torch.tensor(loaded_model.f(*arg)[0].numpy())
      self.assertTrue(torch.allclose(pt_res, tf_res, atol=1e-4))


if __name__ == "__main__":
  base_test_util.main()



"""`torch.distributed` backend implemented with JAX collective ops.

EXPERIMENTAL: This module is still highly experimental, and it may be removed
before any stable release.

Note: JAX collective ops require that axis names be defined in `pmap` or
`shmap`. The distributed backend only supports one axis, named `torch_dist`.
This name is defined by our mirror implementation of `spawn`.
"""

import datetime
import functools
import logging
import os
from typing import List, Optional, Union

import jax
import numpy as np
import torch
import torch.distributed as dist
import torch.distributed._functional_collectives
from torch._C._distributed_c10d import ProcessGroup  # type: ignore
import torch.distributed
import torchax
from jax.sharding import NamedSharding
from jax.sharding import Mesh, PartitionSpec as P
from jax.experimental import mesh_utils
import torch.utils._pytree as torch_pytree
from torchax import interop


class ProcessGroupJax(ProcessGroup):
  """Distributed backend implemented with JAX."""

  def __init__(self, prefix_store, rank, size, timeout):
    super().__init__(rank, size)
    self._group_name = None

  def getBackendName(self):
    return "jax"

  # TODO(wcromar): why doesn't default group name setter work?
  # https://github.com/pytorch/pytorch/blob/7b1988f9222f3dec5cc2012afce84218199748ae/torch/csrc/distributed/c10d/ProcessGroup.cpp#L148-L152
  def _set_group_name(self, name: str) -> None:
    self._group_name = name

  @property
  def group_name(self):
    assert self._group_name
    return self._group_name

  @staticmethod
  def _work(
      tensors: Union[torch.Tensor, List[torch.Tensor],
                     List[List[torch.Tensor]]],
  ) -> dist.Work:
    fut = torch.futures.Future()
    fut.set_result(tensors)
    return torch._C._distributed_c10d._create_work_from_future(fut)

  def _allgather_base(
      self,
      output: torch.Tensor,
      input: torch.Tensor,
      opts=...,
  ) -> dist.Work:
    assert isinstance(input, torchax.tensor.Tensor)
    assert isinstance(output, torchax.tensor.Tensor)
    torch.distributed._functional_collectives.all_gather_tensor_inplace(
        output, input, group=self)
    return self._work(output)

  def allreduce(
      self,
      tensors: List[torch.Tensor],
      opts: dist.AllreduceOptions = ...,
  ) -> dist.Work:
    assert len(tensors) == 1
    assert isinstance(tensors[0], torchax.tensor.Tensor)
    torch.distributed._functional_collectives.all_reduce_inplace(
        tensors[0],
        torch.distributed._functional_collectives.REDUCE_OP_TO_STR[
            opts.reduceOp.op],
        self,
    )

    return self._work(tensors)

  def broadcast(
      self,
      tensors: List[torch.Tensor],
      opts: dist.BroadcastOptions = ...,
  ) -> dist.Work:
    assert len(tensors) == 1
    assert isinstance(tensors[0], torchax.tensor.Tensor)
    tensors[0].copy_(
        torch.distributed._functional_collectives.broadcast(
            tensors[0], opts.rootRank, group=self))

    return self._work(tensors)


dist.Backend.register_backend("jax", ProcessGroupJax, devices=["jax"])


def jax_rendezvous_handler(url: str,
                           timeout: datetime.timedelta = ...,
                           **kwargs):
  """Initialize distributed store with JAX process IDs.

  Requires `$MASTER_ADDR` and `$MASTER_PORT`.
  """
  # TODO(wcromar): jax.distributed.initialize(...) for multiprocess on GPU
  # TODO(wcromar): Can we use the XLA coordinator as a Store? This isn't part
  # of their public Python API
  master_ip = os.environ["MASTER_ADDR"]
  master_port = int(os.environ["MASTER_PORT"])
  # TODO(wcromar): Use `torchrun`'s store if available
  store = dist.TCPStore(
      master_ip,
      master_port,
      jax.process_count(),
      is_master=jax.process_index() == 0,
  )

  yield (store, jax.process_index(), jax.process_count())


dist.register_rendezvous_handler("jax", jax_rendezvous_handler)


def spawn(f, args=(), env: Optional[torchax.tensor.Environment] = None):
  """Wrap `f` in a JAX `pmap` with the axis name `torch_dist` defined.
  `f` is expected to take the replica index as a positional argument, similar
  to `torch.multiprocessing.spawn`.
  Note: `spawn` does not actually create parallel processes.
  """
  env = env or torchax.default_env()

  def jax_wrapper(index, jax_args):
    index, args = env.j2t_iso([index, jax_args])
    torch_outputs = f(index, *args)
    return env.t2j_iso(torch_outputs)

  jax_outputs = jax.pmap(
      jax_wrapper, axis_name="torch_dist")(np.arange(jax.device_count()),
                                           env.t2j_iso(args))
  return env.j2t_iso(jax_outputs)


class DistributedDataParallel(torch.nn.Module):
  """Re-implementation of DistributedDataParallel using JAX SPMD.

  Splits inputs along batch dimension (assumed to be 0) across all devices in
  JAX runtime, including remote devices. Each process should load a distinct
  shard of the input data using e.g. DistributedSampler. Each process' shard
  is then further split among the addressable devices (e.g. local TPU chips)
  by `shard_input`.

  Note: since parameters are replicated across addressable devices, inputs
  must also be SPMD sharded using `shard_input` or `replicate_input`.

  Example usage:

  ```
  jax_model = torchax.distributed.DistributedDataParallel(create_model())
  for data, dataloader:
    jax_data = jax_model.shard_input(data)
    jax_output = jax_model(jax_data)
  ```
  """

  def __init__(
      self,
      module: torch.nn.Module,
      env: Optional[torchax.tensor.Environment] = None,
      **kwargs,
  ):
    if kwargs:
      logging.warning(f"Unsupported kwargs {kwargs}")

    super().__init__()
    self._env = env or torchax.default_env()
    self._mesh = Mesh(
        mesh_utils.create_device_mesh((jax.device_count(),)),
        axis_names=("batch",),
    )
    replicated_state = torch_pytree.tree_map_only(
        torch.Tensor,
        lambda t: self._env.j2t_iso(
            jax.device_put(
                self._env.to_xla(t)._elem, NamedSharding(self._mesh, P()))),
        module.state_dict(),
    )
    # TODO: broadcast
    module.load_state_dict(replicated_state, assign=True)
    self._module = module

  def shard_input(self, inp):
    per_process_batch_size = inp.shape[0]  # assumes batch dim is 0
    per_replica_batch_size = per_process_batch_size // jax.local_device_count()
    per_replica_batches = torch.chunk(inp, jax.local_device_count())
    global_batch_size = per_replica_batch_size * jax.device_count()
    global_batch_shape = (global_batch_size,) + inp.shape[1:]

    sharding = NamedSharding(self._mesh, P("batch"))
    return self._env.j2t_iso(
        jax.make_array_from_single_device_arrays(
            global_batch_shape,
            NamedSharding(self._mesh, P("batch")),
            arrays=[
                jax.device_put(self._env.to_xla(batch)._elem, device) for batch,
                device in zip(per_replica_batches, sharding.addressable_devices)
            ],
        ))

  def replicate_input(self, inp):
    return self._env.j2t_iso(
        jax.device_put(inp._elem, NamedSharding(self._mesh, P())))

  def jit_step(self, func):

    @functools.partial(
        interop.jax_jit, kwargs_for_jax_jit={'donate_argnums': 0})
    def _jit_fn(states, args):
      self.load_state_dict(states)
      outputs = func(*args)
      return self.state_dict(), outputs

    @functools.wraps(func)
    def inner(*args):
      jax_states = self.state_dict()
      new_states, outputs = _jit_fn(jax_states, args)
      self.load_state_dict(new_states)
      return outputs

    return inner

  def forward(self, *args):
    with self._env:
      return self._module(*args)



import unittest
import jax
import jax.numpy as jnp
import torchax
from torchax import interop
import torch


class AutocastTest(unittest.TestCase):

  def setUp(self):
    self.env = torchax.default_env()

  def test_auto_cast_ir(self):
    with self.env:
      with torchax.amp.autocast('jax', dtype=torch.bfloat16, env=self.env):
        a = jax.ShapeDtypeStruct((2, 2), jnp.float32)
        b = jax.ShapeDtypeStruct((2, 2), jnp.float32)
        ir_text = jax.jit(interop.jax_view(torch.matmul)).lower(a, b).as_text()
        self.assertIn('tensor<2x2xbf16>', ir_text)

  def test_auto_cast_matmul(self):
    with self.env:
      a = torch.randn(2, 2, device='jax')
      b = torch.randn(2, 2, device='jax')
      with torchax.amp.autocast('jax', dtype=torch.bfloat16, env=self.env):
        c = a @ b

      self.assertEqual(c.dtype, torch.bfloat16)

      with torch.autocast('cpu', dtype=torch.bfloat16):
        c_cpu = a.cpu() @ b.cpu()

      self.assertTrue(torch.allclose(c.cpu(), c_cpu))


if __name__ == '__main__':
  unittest.main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.

# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
from dataclasses import dataclass
from typing import Optional

import torch
import torch.nn as nn
from torch import Tensor
from torch.nn import functional as F


def find_multiple(n: int, k: int) -> int:
  if n % k == 0:
    return n
  return n + k - (n % k)


@dataclass
class ModelArgs:
  block_size: int = 2048
  vocab_size: int = 32000
  n_layer: int = 32
  n_head: int = 32
  dim: int = 4096
  intermediate_size: int = None
  n_local_heads: int = -1
  head_dim: int = 64
  rope_base: float = 10000
  norm_eps: float = 1e-5
  num_experts: int = 8
  num_activated_experts: int = 2

  def __post_init__(self):
    if self.n_local_heads == -1:
      self.n_local_heads = self.n_head
    if self.intermediate_size is None:
      hidden_dim = 4 * self.dim
      n_hidden = int(2 * hidden_dim / 3)
      self.intermediate_size = find_multiple(n_hidden, 256)
    self.head_dim = self.dim // self.n_head

  @classmethod
  def from_name(cls, name: str):
    if name in transformer_configs:
      return cls(**transformer_configs[name])
    # fuzzy search
    config = [
        config for config in transformer_configs
        if config in str(name).upper() or config in str(name)
    ]
    assert len(config) == 1, name
    return cls(**transformer_configs[config[0]])


transformer_configs = {
    "Mixtral-8x7B-v0.1":
        dict(
            block_size=32768,
            n_layer=32,
            n_head=32,
            n_local_heads=8,
            dim=4096,
            intermediate_size=14336,
            rope_base=1000000.0,
            num_experts=8,
            num_activated_experts=2),
}


class KVCache(nn.Module):

  def __init__(self,
               max_batch_size,
               max_seq_length,
               n_heads,
               head_dim,
               dtype=torch.bfloat16):
    super().__init__()
    cache_shape = (max_batch_size, n_heads, max_seq_length, head_dim)
    self.register_buffer('k_cache', torch.zeros(cache_shape, dtype=dtype))
    self.register_buffer('v_cache', torch.zeros(cache_shape, dtype=dtype))

  def update(self, input_pos, k_val, v_val):
    # input_pos: [S], k_val: [B, H, S, D]
    assert input_pos.shape[0] == k_val.shape[2]

    k_out = self.k_cache
    v_out = self.v_cache
    k_out[:, :, input_pos] = k_val
    v_out[:, :, input_pos] = v_val

    return k_out, v_out


class Transformer(nn.Module):

  def __init__(self, config: ModelArgs) -> None:
    super().__init__()
    self.config = config

    self.tok_embeddings = nn.Embedding(config.vocab_size, config.dim)
    self.layers = nn.ModuleList(
        TransformerBlock(config) for _ in range(config.n_layer))
    self.norm = RMSNorm(config.dim, eps=config.norm_eps)
    self.output = nn.Linear(config.dim, config.vocab_size, bias=False)

    self.freqs_cis: Optional[Tensor] = None
    self.mask_cache: Optional[Tensor] = None
    self.max_batch_size = -1
    self.max_seq_length = -1

  def setup_caches(self, max_batch_size, max_seq_length):
    if self.max_seq_length >= max_seq_length and self.max_batch_size >= max_batch_size:
      return
    head_dim = self.config.dim // self.config.n_head
    max_seq_length = find_multiple(max_seq_length, 8)
    self.max_seq_length = max_seq_length
    self.max_batch_size = max_batch_size
    for b in self.layers:
      b.attention.kv_cache = KVCache(max_batch_size, max_seq_length,
                                     self.config.n_local_heads, head_dim)

    self.freqs_cis = precompute_freqs_cis(self.config.block_size,
                                          self.config.dim // self.config.n_head,
                                          self.config.rope_base)
    self.causal_mask = torch.tril(
        torch.ones(self.max_seq_length, self.max_seq_length, dtype=torch.bool))

  def forward(self, idx: Tensor, input_pos: Optional[Tensor] = None) -> Tensor:
    assert self.freqs_cis is not None, "Caches must be initialized first"
    mask = self.causal_mask[None, None, input_pos]
    freqs_cis = self.freqs_cis[input_pos]
    x = self.tok_embeddings(idx)

    for i, layer in enumerate(self.layers):
      x = layer(x, input_pos, freqs_cis, mask)
    x = self.norm(x)
    logits = self.output(x)
    return logits

  @classmethod
  def from_name(cls, name: str):
    return cls(ModelArgs.from_name(name))


class TransformerBlock(nn.Module):

  def __init__(self, config: ModelArgs) -> None:
    super().__init__()
    self.attention = Attention(config)
    self.block_sparse_moe = MOEFeedForward(config)
    self.ffn_norm = RMSNorm(config.dim, config.norm_eps)
    self.attention_norm = RMSNorm(config.dim, config.norm_eps)

  def forward(self, x: Tensor, input_pos: Tensor, freqs_cis: Tensor,
              mask: Tensor) -> Tensor:
    h = x + self.attention(self.attention_norm(x), freqs_cis, mask, input_pos)
    out = h + self.block_sparse_moe(self.ffn_norm(h))
    return out


class Attention(nn.Module):

  def __init__(self, config: ModelArgs):
    super().__init__()
    assert config.dim % config.n_head == 0

    total_head_dim = (config.n_head +
                      2 * config.n_local_heads) * config.head_dim
    # key, query, value projections for all heads, but in a batch
    self.wqkv = nn.Linear(config.dim, total_head_dim, bias=False)
    self.wo = nn.Linear(config.dim, config.dim, bias=False)
    self.kv_cache = None

    self.n_head = config.n_head
    self.head_dim = config.head_dim
    self.n_local_heads = config.n_local_heads
    self.dim = config.dim
    self._register_load_state_dict_pre_hook(self.load_hook)

  def load_hook(self, state_dict, prefix, *args):
    if prefix + "wq.weight" in state_dict:
      wq = state_dict.pop(prefix + "wq.weight")
      wk = state_dict.pop(prefix + "wk.weight")
      wv = state_dict.pop(prefix + "wv.weight")
      state_dict[prefix + "wqkv.weight"] = torch.cat([wq, wk, wv])

  def forward(self,
              x: Tensor,
              freqs_cis: Tensor,
              mask: Tensor,
              input_pos: Optional[Tensor] = None) -> Tensor:
    bsz, seqlen, _ = x.shape

    kv_size = self.n_local_heads * self.head_dim
    q, k, v = self.wqkv(x).split([self.dim, kv_size, kv_size], dim=-1)

    q = q.view(bsz, seqlen, self.n_head, self.head_dim)
    k = k.view(bsz, seqlen, self.n_local_heads, self.head_dim)
    v = v.view(bsz, seqlen, self.n_local_heads, self.head_dim)

    q = apply_rotary_emb(q, freqs_cis)
    k = apply_rotary_emb(k, freqs_cis)

    q, k, v = map(lambda x: x.transpose(1, 2), (q, k, v))

    if self.kv_cache is not None:
      k, v = self.kv_cache.update(input_pos, k, v)

    k = k.repeat_interleave(self.n_head // self.n_local_heads, dim=1)
    v = v.repeat_interleave(self.n_head // self.n_local_heads, dim=1)
    y = F.scaled_dot_product_attention(q, k, v, attn_mask=mask, dropout_p=0.0)

    y = y.transpose(1, 2).contiguous().view(bsz, seqlen, self.dim)

    y = self.wo(y)
    return y


class ConditionalFeedForward(nn.Module):

  def __init__(self, config):
    super().__init__()
    self.w1 = nn.Parameter(
        torch.empty(config.num_experts, config.intermediate_size, config.dim))
    self.w2 = nn.Parameter(
        torch.empty(config.num_experts, config.dim, config.intermediate_size))
    self.w3 = nn.Parameter(
        torch.empty(config.num_experts, config.intermediate_size, config.dim))

  def forward(self, x: Tensor, expert_indices: Tensor) -> Tensor:
    w1_weights = self.w1[expert_indices]  # [T, A, D, D]
    w3_weights = self.w3[expert_indices]  # [T, A, D, D]
    w2_weights = self.w2[expert_indices]  # [T, A, D, D]
    x1 = F.silu(torch.einsum('ti,taoi -> tao', x, w1_weights))
    x3 = torch.einsum('ti, taoi -> tao', x, w3_weights)
    expert_outs = torch.einsum('tao, taio -> tai', (x1 * x3), w2_weights)
    return expert_outs


class MOEFeedForward(nn.Module):

  def __init__(self, config) -> None:
    super().__init__()
    self.gate = nn.Linear(config.dim, config.num_experts, bias=False)
    self.cond_ffn = ConditionalFeedForward(config)
    self.dim = config.dim
    self.num_activated_experts = config.num_activated_experts

  def forward(self, x: Tensor) -> Tensor:
    x = x.view(-1, self.dim)
    # T = num_tokens, E = num_experts, D = hidden dim, A = activated experts
    # x: [T, D]
    scores = self.gate(x)  # [T, E]
    expert_weights = F.softmax(scores, dim=-1)
    expert_weights, expert_indices = torch.topk(
        expert_weights, self.num_activated_experts, dim=-1)  # [T, A], [T, A]
    expert_weights /= expert_weights.sum(dim=-1, keepdim=True)  # [T, A]
    expert_outs = self.cond_ffn(x, expert_indices)
    return torch.einsum('tai,ta -> ti', expert_outs, expert_weights)


class RMSNorm(nn.Module):

  def __init__(self, dim: int, eps: float = 1e-5):
    super().__init__()
    self.eps = eps
    self.weight = nn.Parameter(torch.ones(dim))

  def _norm(self, x):
    return x * torch.rsqrt(torch.mean(x * x, dim=-1, keepdim=True) + self.eps)

  def forward(self, x: Tensor) -> Tensor:
    output = self._norm(x.float()).type_as(x)
    return output * self.weight


def precompute_freqs_cis(seq_len: int,
                         n_elem: int,
                         base: int = 10000) -> Tensor:
  freqs = 1.0 / (
      base**(torch.arange(0, n_elem, 2)[:(n_elem // 2)].float() / n_elem))
  t = torch.arange(seq_len, device=freqs.device)
  freqs = torch.outer(t, freqs)
  freqs_cis = torch.polar(torch.ones_like(freqs), freqs)
  cache = torch.stack([freqs_cis.real, freqs_cis.imag], dim=-1)
  return cache.to(dtype=torch.bfloat16)


def apply_rotary_emb(x: Tensor, freqs_cis: Tensor) -> Tensor:
  xshaped = x.float().reshape(*x.shape[:-1], -1, 2)
  freqs_cis = freqs_cis.view(1, xshaped.size(1), 1, xshaped.size(3), 2)
  x_out2 = torch.stack(
      [
          xshaped[..., 0] * freqs_cis[..., 0] -
          xshaped[..., 1] * freqs_cis[..., 1],
          xshaped[..., 1] * freqs_cis[..., 0] +
          xshaped[..., 0] * freqs_cis[..., 1],
      ],
      -1,
  )

  x_out2 = x_out2.flatten(3)
  return x_out2.type_as(x)



import functools
import torch
import unittest
import torchax
from torchax import interop, jax_device
import torchax
import jax


def is_tpu_available():
  """Checks if any TPU devices are available to JAX."""
  try:
    # jax.devices('tpu') will return a list of TPU devices if available.
    # If no TPUs are found or JAX is not configured for TPU,
    # it will raise a RuntimeError.
    tpu_devices = jax.devices('tpu')
    return len(tpu_devices) > 0
  except RuntimeError:
    return False


class InteropTest(unittest.TestCase):

  def setUp(self):
    torchax.enable_globally()

  def test_mod_attr(self):

    class Child(torch.nn.Module):

      def __init__(self):
        super().__init__()
        self.x = torch.ones(10, 10)

    class ModuleWithUnregisteredTensor(torch.nn.Module):

      def __init__(self):
        super().__init__()
        self.a = torch.nn.Linear(100, 100)
        self.b = torch.nn.Parameter(torch.ones(10, 10))
        c = torch.ones(10, 10)
        self.register_buffer('c', c)
        self.register_buffer('c2', c, persistent=False)
        self.d = torch.ones(10, 10)
        self.m1 = Child()

    m = ModuleWithUnregisteredTensor()
    params, buffers = interop.extract_all_buffers(m)
    self.assertEqual(set(params.keys()), {'a.weight', 'a.bias', 'b'})
    self.assertEqual(set(buffers.keys()), {'c', 'c2', 'd', 'm1.x'})

    interop.set_all_buffers(m, {'a.weight': torch.tensor([0.0])},
                            {'m1.x': torch.tensor([0.0])})
    self.assertEqual(m.a.weight.item(), 0)
    self.assertEqual(m.m1.x.item(), 0)

  def test_j2t_autograd_forward(self):
    with torchax.default_env():
      # Setup
      def fn(x):
        return x + 1

      j2t_fn = interop.j2t_autograd(fn)
      x = torch.ones(2, 2, requires_grad=True, device='jax')

      # Act
      actual = j2t_fn(x)

      # Assert
      expected = torch.ones(2, 2) + 1
      torch.testing.assert_close(actual, expected, check_device=False)

  def test_j2t_autograd_backward(self):
    with torchax.default_env():
      # Setup
      def fn(x):
        return x * 2

      j2t_fn = interop.j2t_autograd(fn)
      x = torch.ones(2, 2, device='jax').requires_grad_()

      # Act
      actual = j2t_fn(x)
      actual.sum().backward()

      # Assert
      expected = torch.ones(2, 2) * 2
      torch.testing.assert_close(x.grad, expected, check_device=False)

  def test_module_with_shared_weights(self):

    # arrange
    class ModuleWithSharedWeights(torch.nn.Module):

      def __init__(self):
        super().__init__()
        self.a = torch.nn.Linear(10, 10)
        self.b = self.a

      def forward(self, x):
        return self.a(self.b(x))

    m = ModuleWithSharedWeights().to('jax')

    m_jitted = interop.JittableModule(m, dedup_parameters=True)

    # a's weights and bias and b's weights and bias
    self.assertEqual(len(m.state_dict()), 4)

    # b's weights and bias are deduped
    self.assertEqual(len(m_jitted.params), 2)
    x = torch.randn(10, 10).to('jax')
    expected = m(x)

    # act
    actual = m_jitted(x)

    # assert
    torch.testing.assert_allclose(actual, expected)

    # arrange
    # make sure buffer donation works
    functional_forward = interop.jax_jit(
        functools.partial(m_jitted.functional_call, 'forward'),
        kwargs_for_jax_jit={'donate_argnums': (0,)})

    # act
    actual = functional_forward(m_jitted.params, m_jitted.buffers, x)
    # assert
    torch.testing.assert_allclose(actual, expected)

  def test_to_jax_device(self):
    a = torch.ones(3, 3)

    if is_tpu_available():
      # by default if tpu is available, to jax will be to tpu
      e = a.to("jax")
      self.assertEqual(e.jax_device.platform, "tpu")
      self.assertEqual(e.device.type, "jax")
    else:
      e = a.to("jax")
      self.assertEqual(e.jax_device.platform, "cpu")
      self.assertEqual(e.device.type, "jax")

    with jax_device("cpu"):
      # move torch.tensor to torchax.tensor CPU
      b = a.to("jax")
      self.assertEqual(b.jax_device.platform, "cpu")
      self.assertEqual(b.device.type, "jax")

    if is_tpu_available():
      # move torch.tensor to torchax.tensor TPU
      with jax_device("tpu"):
        c = a.to("jax")
        self.assertEqual(c.jax_device.platform, "tpu")
        self.assertEqual(c.device.type, "jax")

      # move torchax.tensor on CPU to TPU
      with jax_device("tpu"):
        self.assertEqual(b.jax_device.platform, "cpu")
        self.assertEqual(c.device.type, "jax")
        c = b.to("jax")
        self.assertEqual(c.jax_device.platform, "tpu")
        self.assertEqual(c.device.type, "jax")

      # move torchax.tensor on TPU to CPU
      with jax_device("cpu"):
        self.assertEqual(c.jax_device.platform, "tpu")
        self.assertEqual(c.device.type, "jax")
        d = c.to("jax")
        self.assertEqual(d.jax_device.platform, "cpu")
        self.assertEqual(d.device.type, "jax")


if __name__ == '__main__':
  unittest.main()



"""
This is the script from this tutorial:
https://pytorch.org/tutorials/beginner/introyt/trainingyt.html
"""

import functools
from torchax import train, interop
import torch
from torch.utils import _pytree as pytree
import torchvision
import torchvision.transforms as transforms
import torchax
import torchax.interop
import jax
import optax
import numpy as np

# PyTorch TensorBoard support
from torch.utils.tensorboard import SummaryWriter
from datetime import datetime

env = torchax.enable_globally()

transform = transforms.Compose(
    [transforms.ToTensor(),
     transforms.Normalize((0.5,), (0.5,))])

# Create datasets for training & validation, download if necessary
training_set = torchvision.datasets.FashionMNIST(
    './data', train=True, transform=transform, download=True)
validation_set = torchvision.datasets.FashionMNIST(
    './data', train=False, transform=transform, download=True)

# Create data loaders for our datasets; shuffle for training, not for validation
training_loader = torch.utils.data.DataLoader(
    training_set, batch_size=4, shuffle=True)
validation_loader = torch.utils.data.DataLoader(
    validation_set, batch_size=4, shuffle=False)

# Class labels
classes = ('T-shirt/top', 'Trouser', 'Pullover', 'Dress', 'Coat', 'Sandal',
           'Shirt', 'Sneaker', 'Bag', 'Ankle Boot')

# Report split sizes
print('Training set has {} instances'.format(len(training_set)))
print('Validation set has {} instances'.format(len(validation_set)))

import numpy as np
import torch.nn as nn
import torch.nn.functional as F


# PyTorch models inherit from torch.nn.Module
class GarmentClassifier(nn.Module):

  def __init__(self):
    super(GarmentClassifier, self).__init__()
    self.fc1 = nn.Linear(28 * 28, 120)
    self.fc2 = nn.Linear(120, 84)
    self.fc3 = nn.Linear(84, 10)

  def forward(self, x):
    x = x.view(-1, 28 * 28)
    x = F.relu(self.fc1(x))
    x = F.relu(self.fc2(x))
    x = self.fc3(x)
    return x


model = GarmentClassifier()
loss_fn = torch.nn.CrossEntropyLoss()

jax_optimizer = optax.adam(0.01)

model.to('jax')  # move the model to jax device
model_jittable = interop.JittableModule(model)
weights = model_jittable.params  # these are trainable parameters
buffers = model_jittable.buffers  # these are non-trainable parameters

opt_state = interop.call_jax(jax_optimizer.init, weights)
model_fn = functools.partial(model_jittable.functional_call, 'forward')

train_step = train.make_train_step(model_fn, loss_fn, jax_optimizer)

train_step = interop.jax_jit(
    train_step, kwargs_for_jax_jit={'donate_argnums': (0, 2)})

# NB: Loss functions expect data in batches, so we're creating batches of 4
# Represents the model's confidence in each of the 10 classes for a given input
dummy_inputs = torch.rand(4, 28, 28).to('jax')
dummy_outputs = torch.rand(4, 10).to('jax')
# Represents the correct class among the 10 being tested
dummy_labels = torch.tensor([1, 5, 3, 7]).to('jax')

# test train_step


def train_one_epoch(weights, buffers, opt_state, epoch_index, tb_writer):
  running_loss = 0.
  last_loss = 0.

  # Here, we use enumerate(training_loader) instead of
  # iter(training_loader) so that we can track the batch
  # index and do some intra-epoch reporting
  for i, data in enumerate(training_loader):
    inputs, labels = data

    inputs = inputs.to('jax')
    labels = labels.to('jax')

    loss, weights, opt_state = train_step(weights, buffers, opt_state, inputs,
                                          labels)

    # Gather data and report
    running_loss += loss.item()
    if i % 1000 == 999:
      last_loss = running_loss / 1000  # loss per batch
      print('  batch {} loss: {}'.format(i + 1, last_loss))
      tb_x = epoch_index * len(training_loader) + i + 1
      tb_writer.add_scalar('Loss/train', last_loss, tb_x)
      running_loss = 0.

  return last_loss, weights, opt_state


# Initializing in a separate cell so we can easily add more epochs to the same run
timestamp = datetime.now().strftime('%Y%m%d_%H%M%S')
writer = SummaryWriter('runs/fashion_trainer_{}'.format(timestamp))
epoch_number = 0
EPOCHS = 2
best_vloss = 1_000_000.

for epoch in range(EPOCHS):
  print('EPOCH {}:'.format(epoch_number + 1))

  avg_loss, weights, opt_state = train_one_epoch(weights, buffers, opt_state,
                                                 epoch_number, writer)
  print(avg_loss)



import jax
import jax.numpy as jnp
from litgpt import config
from litgpt import model
from litgpt.data import Alpaca
from litgpt.tokenizer import Tokenizer
import lightning
import torch
from collections import defaultdict
from jax.experimental import shard_map

import torch.nn.functional
import torchax.interop

from . import utils
from . import model as editted_model
import os


def _setup_default_env():
  os.environ.setdefault('TF_CPP_MIN_LOG_LEVEL', '1')
  os.environ.setdefault('GRPC_VERBOSITY', 'ERROR')
  os.environ.setdefault('ALLOW_MULTIPLE_LIBTPU_LOAD', '1')
  # only need for tpu v4
  # os.environ.setdefault('TPU_MEGACORE', 'megacore_dense')
  tpu_args = "--xla_tpu_enable_async_collective_fusion_fuse_all_gather=true --xla_tpu_megacore_fusion_allow_ags=false --xla_enable_async_collective_permute=true --xla_tpu_enable_ag_backward_pipelining=true --xla_tpu_enable_data_parallel_all_reduce_opt=true --xla_tpu_data_parallel_opt_different_sized_ops=true --xla_tpu_enable_async_collective_fusion=true --xla_tpu_enable_async_collective_fusion_multiple_steps=true --xla_tpu_overlap_compute_collective_tc=true --xla_enable_async_all_gather=true"

  os.environ.setdefault('LIBTPU_INIT_ARGS', tpu_args)


_setup_default_env()

default_checkpoint_dir = '/home/hanq/litgpt/checkpoints/meta-llama/Meta-Llama-3-8B/'


class GPTLightningModule(lightning.LightningModule):

  def __init__(self, gpt):
    super().__init__()
    self.gpt = utils.FSDPv2(gpt)

  def training_step(self, batch, batch_idx):
    x, y = batch
    logits = self.gpt.forward(x)
    num_tokens = logits.shape[-1]
    logits = logits[..., :-1, :].reshape(-1, num_tokens)
    y = y[..., 1:].reshape(-1)
    return torch.nn.functional.cross_entropy(logits, y)

  def configure_optimizers(self):
    return None


from jax.experimental import mesh_utils

P = jax.sharding.PartitionSpec
mesh = jax.sharding.Mesh(
    mesh_utils.create_device_mesh(utils.num_partitions),
    axis_names=utils.global_axis,
)


class GPTOutline(torch.nn.Module):

  def __init__(self, gpt_orig):
    super().__init__()
    self.gpt_orig = gpt_orig

    def one_layer(weights, args):
      return torch.func.functional_call(self.gpt_orig.transformer.h[0], weights,
                                        args)

    self.one_layer = torchax.interop.jax_jit(one_layer)

  def forward(self, idx: torch.Tensor, input_pos=None) -> torch.Tensor:
    T = idx.size(1)
    if self.gpt_orig.max_seq_length < T:
      raise ValueError(
          f"Cannot forward sequence of length {T}, max seq length is only {self.max_seq_length}."
      )

    cos = self.gpt_orig.cos[:T]
    sin = self.gpt_orig.sin[:T]
    mask = None

    x = self.gpt_orig.transformer.wte(
        idx)  # token embeddings of shape (b, t, n_embd)
    editted_model.reapply_sharding(x)
    if self.gpt_orig.config.scale_embeddings:
      x = x * (self.gpt_orig.config.n_embd**0.5)

    for block in self.gpt_orig.transformer.h:
      args = (x, cos, sin, mask, input_pos)
      weights = block.state_dict()
      x = self.one_layer(weights, args)
      editted_model.reapply_sharding(x)

    x = self.gpt_orig.transformer.ln_f(x)
    editted_model.reapply_sharding(x)
    res = self.gpt_orig.lm_head(x)  # (b, t, vocab_size)
    editted_model.reapply_sharding(res)
    return res


class GPTFori:

  def __init__(self, gpt_orig, manual_all_gather=False):
    super().__init__()
    self.gpt_orig = gpt_orig

    one_block = self.gpt_orig.transformer.h[0]
    self.manual_all_gather = manual_all_gather

    def one_layer(args, weights):
      # inputs are jax array
      orig_args = args

      x, cos, sin, mask, input_pos = args
      if self.manual_all_gather:
        weights, cos, sin = jax.lax.all_gather((weights, cos, sin),
                                               'fsdp',
                                               tiled=True)
      args = (x, cos, sin, mask, input_pos)
      args, weights = torchax.default_env().j2t_iso((args, weights))
      res = torch.func.functional_call(one_block, weights, args)
      res = torchax.default_env().t2j_iso(res)
      return (res, *orig_args[1:]), jnp.array([0])

    if self.manual_all_gather:
      one_layer = shard_map.shard_map(
          one_layer,
          mesh=mesh,
          in_specs=(P(*utils.global_axis), P(*utils.global_axis)),
          out_specs=(P(*utils.global_axis), P()),
          check_rep=False)

    one_layer = jax.checkpoint(
        one_layer,
        policy=jax.checkpoint_policies.dots_with_no_batch_dims_saveable)

    def compiled_block(weights, x):
      x, _ = jax.lax.scan(one_layer, x, weights, unroll=4)
      return x[0]

    self.compiled_block = compiled_block

    self.weights = {
        'wte': self.gpt_orig.transformer.wte.state_dict(),
        'block': self.make_weights_scan(),
        'ln_f': self.gpt_orig.transformer.ln_f.state_dict(),
        'lm_head': self.gpt_orig.lm_head.state_dict(),
        'sin': self.gpt_orig.sin,
        'cos': self.gpt_orig.cos,
    }

  def make_weights_scan(self):
    temp = defaultdict(list)  # key to list of tensors
    for block in self.gpt_orig.transformer.h:
      state_dict = block.state_dict()
      for k, v in state_dict.items():
        temp[k].append(v)

    temp = {k: torch.stack(v) for k, v in temp.items()}
    return temp

  def forward_with_weights(self, weights, idx):
    T = idx.size(1)
    if self.gpt_orig.max_seq_length < T:
      raise ValueError(
          f"Cannot forward sequence of length {T}, max seq length is only {self.max_seq_length}."
      )

    cos = weights['cos'][:T]
    sin = weights['sin'][:T]
    mask = None

    x = torch.func.functional_call(
        self.gpt_orig.transformer.wte,
        weights['wte'],
        idx,
    )

    editted_model.reapply_sharding(x)

    if self.gpt_orig.config.scale_embeddings:
      x = x * (self.config.n_embd**0.5)
    editted_model.reapply_sharding(x)

    args = (x, cos, sin, mask, None)
    #import pdb; pdb.set_trace()
    x = torchax.interop.call_jax(
        self.compiled_block,
        weights['block'],
        args,
    )
    editted_model.reapply_sharding(x)

    x = torch.func.functional_call(self.gpt_orig.transformer.ln_f,
                                   weights['ln_f'], x)
    editted_model.reapply_sharding(x)
    x = torch.func.functional_call(self.gpt_orig.lm_head, weights['lm_head'], x)
    editted_model.reapply_sharding(x)
    return x


import logging
import torchax

# Modes:

REGULAR = 'regular'
OUTLINED_LAYER = 'outlined'  # jax.jit a block to get faster compilation
SCAN_LAYER = 'scan_layer'  # jax.lax.scan for looping layers to get faster compilation
SCAN_LAYER_MANUAL = 'scan_manual'  # jax.lax.scan AND shmap layers to get faster compilation


def main_one(
    use_flash_attention=True,
    seqlen=8196,
    n_layers=32,
    batch_size=8,
    checkpoint_dir=default_checkpoint_dir,
    mode='regular',
    use_editted_model=False,
):
  logging.getLogger("jax").setLevel(logging.DEBUG)
  print(f"Running with parameters {locals()}")
  utils.SEQLEN = seqlen
  utils.BATCH = batch_size
  env = torchax.default_env()
  env.config.use_tpu_flash_attention = use_flash_attention
  cfg = config.Config.from_name("Meta-Llama-3-8B")
  cfg.n_layer = n_layers
  #cfg.n_layer = 32
  if use_editted_model:
    gpt = editted_model.GPT(cfg)
  else:
    gpt = model.GPT(cfg)
  gpt.to(torch.bfloat16)

  env.config.shmap_flash_attention = mode != SCAN_LAYER_MANUAL
  use_fori = False

  if mode in (SCAN_LAYER, SCAN_LAYER_MANUAL):
    gpt = GPTFori(gpt, mode == SCAN_LAYER_MANUAL)
    use_fori = True
  elif mode == OUTLINED_LAYER:
    gpt = GPTOutline(gpt)

  light_mod = GPTLightningModule(gpt)
  tokenizer = Tokenizer(checkpoint_dir)
  data = Alpaca(num_workers=1)
  data.connect(
      tokenizer=tokenizer, batch_size=batch_size, max_seq_length=utils.SEQLEN)
  data.prepare_data()
  data.setup()
  train_loader = data.train_dataloader()

  with mesh:
    trainer = utils.JaxTrainer(use_fori)
    if use_fori:
      return trainer.fit_model_fori(gpt, train_loader)
    else:
      return trainer.fit(light_mod, train_loader)


def main(
    use_flash_attention=True,
    seqlen=8196,
    n_layers=32,
    batch_size=8,
    checkpoint_dir=default_checkpoint_dir,
    mode='regular',
    use_editted_model=False,
):
  if mode == 'all':
    from jaxlib.xla_extension import XlaRuntimeError
    res = []
    for m, editted in ((SCAN_LAYER, False), (SCAN_LAYER_MANUAL, False),
                       (REGULAR, False), (REGULAR, True), (OUTLINED_LAYER,
                                                           True)):
      try:
        run_time, comp_time = main_one(
            use_flash_attention,
            seqlen,
            n_layers,
            batch_size,
            checkpoint_dir,
            m,
            use_editted_model=editted,
        )
        res.append((m, editted, run_time, comp_time))
      except XlaRuntimeError as e:
        import traceback
        traceback.print_exc()
        res.append((m, editted, 'OOM', ''))
    for m, e, r, c in res:
      print(f'{m}-edit={e}: \t {r} \t {c} ')

  else:
    main_one(use_flash_attention, seqlen, n_layers, batch_size, checkpoint_dir,
             mode, use_editted_model)


if __name__ == '__main__':
  import fire
  fire.Fire(main)



# Copyright Lightning AI. Licensed under the Apache License 2.0, see LICENSE file.
"""Full definition of a decoder-only transformer-based language model, all of it in this single file.

Based on the nanoGPT implementation: https://github.com/karpathy/nanoGPT and
https://github.com/EleutherAI/gpt-neox/tree/main/megatron/model.
"""
import jax
from jax.sharding import PartitionSpec as P
import math
from typing import Any, Optional, Tuple

import torch
import torch.nn as nn
from typing_extensions import Self

from litgpt.config import Config


def reapply_sharding(x):
  x._elem = jax.lax.with_sharding_constraint(x._elem, P('fsdp'))


class GPT(nn.Module):

  def __init__(self, config: Config, use_fori_loop=False) -> None:
    super().__init__()
    assert config.padded_vocab_size is not None
    self.config = config

    self.lm_head = nn.Linear(
        config.n_embd, config.padded_vocab_size, bias=config.lm_head_bias)
    self.transformer = nn.ModuleDict(
        dict(
            wte=nn.Embedding(config.padded_vocab_size, config.n_embd),
            h=nn.ModuleList(Block(config) for _ in range(config.n_layer)),
            ln_f=config.norm_class(config.n_embd, eps=config.norm_eps),
        ))
    self.max_seq_length = self.config.block_size
    self.mask_cache: Optional[torch.Tensor] = None

  @property
  def max_seq_length(self) -> int:
    return self._max_seq_length

  @max_seq_length.setter
  def max_seq_length(self, value: int) -> None:
    """
        When doing inference, the sequences used might be shorter than the model's context length.
        This allows setting a smaller number to avoid allocating unused memory
        """
    if value > self.config.block_size:
      raise ValueError(
          f"Cannot attend to {value}, block size is only {self.config.block_size}"
      )
    self._max_seq_length = value
    if not hasattr(self, "cos"):
      # first call
      cos, sin = self.rope_cache()
      self.register_buffer("cos", cos, persistent=False)
      self.register_buffer("sin", sin, persistent=False)
    # override
    elif value != self.cos.size(0):
      self.cos, self.sin = self.rope_cache(device=self.cos.device)
    # the mask and kv cache size will get updated on `set_kv_cache`. we cannot update it here because we don't know
    # if the kv cache is expected

  def reset_parameters(self) -> None:
    # Trigger resetting the rope-cache
    self.cos, self.sin = self.rope_cache(device=self.cos.device)

  def _init_weights(self, module: nn.Module) -> None:
    """Meant to be used with `gpt.apply(gpt._init_weights)`."""
    if isinstance(module, nn.Linear):
      torch.nn.init.normal_(module.weight, mean=0.0, std=0.02)
      if module.bias is not None:
        torch.nn.init.zeros_(module.bias)
    elif isinstance(module, nn.Embedding):
      torch.nn.init.normal_(module.weight, mean=0.0, std=0.02)

  def forward(self,
              idx: torch.Tensor,
              input_pos: Optional[torch.Tensor] = None) -> torch.Tensor:
    T = idx.size(1)
    if self.max_seq_length < T:
      raise ValueError(
          f"Cannot forward sequence of length {T}, max seq length is only {self.max_seq_length}."
      )

    if input_pos is not None:  # use the kv cache
      cos = self.cos.index_select(0, input_pos)
      sin = self.sin.index_select(0, input_pos)
      if self.mask_cache is None:
        raise TypeError("You need to call `gpt.set_kv_cache()`")
      mask = self.mask_cache.index_select(2, input_pos)
    else:
      cos = self.cos[:T]
      sin = self.sin[:T]
      mask = None
    x = self.transformer.wte(idx)  # token embeddings of shape (b, t, n_embd)
    reapply_sharding(x)
    if self.config.scale_embeddings:
      x = x * (self.config.n_embd**0.5)

    for block in self.transformer.h:
      x = block(x, cos, sin, mask, input_pos)
      reapply_sharding(x)
    x = self.transformer.ln_f(x)
    res = self.lm_head(x)  # (b, t, vocab_size)
    reapply_sharding(res)
    return res

  @classmethod
  def from_name(cls, name: str, **kwargs: Any) -> Self:
    return cls(Config.from_name(name, **kwargs))

  def rope_cache(
      self,
      device: Optional[torch.device] = None
  ) -> Tuple[torch.Tensor, torch.Tensor]:
    return build_rope_cache(
        seq_len=self.max_seq_length,
        n_elem=self.config.rope_n_elem,
        device=device,
        condense_ratio=self.config.rope_condense_ratio,
        base=self.config.rope_base,
    )

  def set_kv_cache(
      self,
      batch_size: int,
      rope_cache_length: Optional[int] = None,
      device: Optional[torch.device] = None,
      dtype: Optional[torch.dtype] = None,
  ) -> None:
    if rope_cache_length is None:
      rope_cache_length = self.cos.size(-1)
    max_seq_length = self.max_seq_length

    # initialize the kv cache for all blocks
    for block in self.transformer.h:
      block.attn.kv_cache = block.attn.build_kv_cache(batch_size,
                                                      max_seq_length,
                                                      rope_cache_length, device,
                                                      dtype)

    if self.mask_cache is None or self.mask_cache.size(3) != max_seq_length:
      # passing `attn_mask` to SDPA disables the flash implementation. since we only need the mask
      # for the kv-cache support (only during inference), we only create it in that situation
      self.mask_cache = build_mask_cache(max_seq_length, device)

  def clear_kv_cache(self) -> None:
    self.mask_cache = None
    for block in self.transformer.h:
      block.attn.kv_cache = None


class Block(nn.Module):

  def __init__(self, config: Config) -> None:
    super().__init__()
    if not config.parallel_residual and config.shared_attention_norm:
      raise NotImplementedError(
          "No checkpoint amongst the ones we support uses this configuration"
          " (non-parallel residual and shared attention norm).")

    self.norm_1 = config.norm_class(config.n_embd, eps=config.norm_eps)
    self.attn = CausalSelfAttention(config)
    self.norm_2 = None if config.shared_attention_norm else config.norm_class(
        config.n_embd, eps=config.norm_eps)
    self.mlp = config.mlp_class(config)

    self.config = config

  def forward(
      self,
      x: torch.Tensor,
      cos: torch.Tensor,
      sin: torch.Tensor,
      mask: Optional[torch.Tensor] = None,
      input_pos: Optional[torch.Tensor] = None,
  ) -> torch.Tensor:
    """
        Non-parallel residual       Parallel residual
           ┌─ x                     ┌─ x ────────────┐             Note: if `shared_attention_norm` is True,
           │  ↓                     │  ↓             ↓                   the output from `norm_1` is reused
           │  norm_1                │  norm_1  ───►  norm_2
           │  ↓                     │  ↓             ↓
           │  attn                  │  attn          mlp
           │  ↓                     │  ↓             │
        ┌─ └► +                     └► + ◄───────────┘
        │     norm_2
        │     ↓
        │     mlp
        │     ↓
        └───► +
        """

    x_normed = self.norm_1(x)
    attention_output = self.attn(x_normed, cos, sin, mask, input_pos)

    if self.config.parallel_residual:
      x_normed = x_normed if self.config.shared_attention_norm else self.norm_2(
          x)
      x = self.mlp(x_normed) + attention_output + x
    else:
      x = attention_output + x
      x = self.mlp(self.norm_2(x)) + x
    return x


class CausalSelfAttention(nn.Module):

  def __init__(self, config: Config) -> None:
    super().__init__()
    shape = (config.n_head + 2 * config.n_query_groups) * config.head_size
    # key, query, value projections for all heads, but in a batch
    self.attn = nn.Linear(config.n_embd, shape, bias=config.bias)
    # output projection
    # if `head_size` is explicitly specified in the config, `n_emd` might not be equal to `head_size * n_head`
    self.proj = nn.Linear(
        config.head_size * config.n_head, config.n_embd, bias=config.bias)
    # disabled by default
    self.kv_cache: Optional[KVCache] = None

    self.config = config

  def forward(
      self,
      x: torch.Tensor,
      cos: torch.Tensor,
      sin: torch.Tensor,
      mask: Optional[torch.Tensor] = None,
      input_pos: Optional[torch.Tensor] = None,
  ) -> torch.Tensor:
    B, T, C = x.size(
    )  # batch size, sequence length, embedding dimensionality (n_embd)

    qkv = self.attn(x)
    qkv._elem = jax.lax.with_sharding_constraint(qkv._elem, P('fsdp'))

    # assemble into a number of query groups to support MHA, MQA and GQA together (see `config.n_query_groups`)
    q_per_kv = self.config.n_head // self.config.n_query_groups
    total_qkv = q_per_kv + 2  # each group has 1+ queries, 1 key, and 1 value
    qkv = qkv.view(B, T, self.config.n_query_groups, total_qkv,
                   self.config.head_size)
    qkv = qkv.permute(0, 2, 3, 1, 4)  # (B, n_query_groups, total_qkv, T, hs)

    # split batched computation into three
    q, k, v = qkv.split((q_per_kv, 1, 1), dim=2)

    # maybe repeat k and v if for the non multi-head attention cases
    # training: flash attention requires it
    # inference: multi-query would require a full kv cache so avoid it to limit its memory usage
    if self.config.n_query_groups != self.config.n_head and (
        input_pos is None or self.config.n_query_groups != 1):
      k = k.expand(B, self.config.n_query_groups, q_per_kv, T,
                   self.config.head_size)
      v = v.expand(B, self.config.n_query_groups, q_per_kv, T,
                   self.config.head_size)

    q = q.reshape(B, -1, T, self.config.head_size)  # (B, nh_q, T, hs)
    k = k.reshape(B, -1, T, self.config.head_size)  # (B, nh_k, T, hs)
    v = v.reshape(B, -1, T, self.config.head_size)  # (B, nh_v, T, hs)

    q_roped = apply_rope(q[..., :self.config.rope_n_elem], cos, sin)
    k_roped = apply_rope(k[..., :self.config.rope_n_elem], cos, sin)
    q = torch.cat((q_roped, q[..., self.config.rope_n_elem:]), dim=-1)
    k = torch.cat((k_roped, k[..., self.config.rope_n_elem:]), dim=-1)

    if input_pos is not None:
      if not isinstance(self.kv_cache, KVCache):
        raise TypeError("You need to call `gpt.set_kv_cache()`")
      k, v = self.kv_cache(input_pos, k, v)

    y = self.scaled_dot_product_attention(q, k, v, mask)

    y = y.reshape(
        B, T, self.config.head_size *
        self.config.n_head)  # re-assemble all head outputs side by side

    # output projection
    return self.proj(y)

  def scaled_dot_product_attention(
      self,
      q: torch.Tensor,
      k: torch.Tensor,
      v: torch.Tensor,
      mask: Optional[torch.Tensor] = None) -> torch.Tensor:
    scale = 1.0 / math.sqrt(self.config.head_size)
    y = torch.nn.functional.scaled_dot_product_attention(
        q,
        k,
        v,
        attn_mask=mask,
        dropout_p=0.0,
        scale=scale,
        is_causal=mask is None)
    return y.transpose(1, 2)

  def build_kv_cache(
      self,
      batch_size: int,
      max_seq_length: int,
      rope_cache_length: Optional[int] = None,
      device: Optional[torch.device] = None,
      dtype: Optional[torch.dtype] = None,
  ) -> "KVCache":
    heads = 1 if self.config.n_query_groups == 1 else self.config.n_head
    v_shape = (batch_size, heads, max_seq_length, self.config.head_size)
    if rope_cache_length is None:
      if self.config.rotary_percentage != 1.0:
        raise TypeError(
            "Please pass the `rope_cache_length=gpt.cos.size(-1)` value")
      k_shape = v_shape
    else:
      k_shape = (
          batch_size,
          heads,
          max_seq_length,
          rope_cache_length + self.config.head_size - self.config.rope_n_elem,
      )
    return KVCache(k_shape, v_shape, device=device, dtype=dtype)


class GptNeoxMLP(nn.Module):

  def __init__(self, config: Config) -> None:
    super().__init__()
    self.fc = nn.Linear(
        config.n_embd, config.intermediate_size, bias=config.bias)
    self.proj = nn.Linear(
        config.intermediate_size, config.n_embd, bias=config.bias)

    self.config = config

  def forward(self, x: torch.Tensor) -> torch.Tensor:
    x = self.fc(x)
    x = torch.nn.functional.gelu(x, approximate=self.config.gelu_approximate)
    return self.proj(x)


class LLaMAMLP(nn.Module):

  def __init__(self, config: Config) -> None:
    super().__init__()
    self.fc_1 = nn.Linear(
        config.n_embd, config.intermediate_size, bias=config.bias)
    self.fc_2 = nn.Linear(
        config.n_embd, config.intermediate_size, bias=config.bias)
    self.proj = nn.Linear(
        config.intermediate_size, config.n_embd, bias=config.bias)

    self.config = config

  def forward(self, x: torch.Tensor) -> torch.Tensor:
    x_fc_1 = self.fc_1(x)
    x_fc_2 = self.fc_2(x)
    x = torch.nn.functional.silu(x_fc_1) * x_fc_2
    return self.proj(x)


class GemmaMLP(LLaMAMLP):

  def forward(self, x: torch.Tensor) -> torch.Tensor:
    x_fc_1 = self.fc_1(x)
    x_fc_2 = self.fc_2(x)
    x = torch.nn.functional.gelu(
        x_fc_1, approximate=self.config.gelu_approximate) * x_fc_2
    return self.proj(x)


class LLaMAMoE(nn.Module):

  def __init__(self, config: Config) -> None:
    super().__init__()
    self.gate = nn.Linear(config.n_embd, config.n_expert, bias=False)
    self.experts = nn.ModuleList(
        LLaMAMLP(config) for _ in range(config.n_expert))

    self.config = config

  def forward(self, x: torch.Tensor) -> torch.Tensor:
    """
        Derived from: https://github.com/mistralai/mistral-src/blob/b46d6/moe_one_file_ref.py#L203-L219
        See also figure 1 in https://arxiv.org/abs/2211.15841
        """
    B, T, C = x.size(
    )  # batch size, sequence length, embedding dimensionality (n_embd)
    x = x.view(-1, C)  # (B*T, C)
    router = self.gate(x)  # (B*T, n_expert)
    probs, indices = torch.topk(
        router, self.config.n_expert_per_token)  # (B*T, n_expert_per_token)
    probs = probs.softmax(dim=1, dtype=torch.float).to(dtype=x.dtype)
    masks = indices.unsqueeze(-1) == torch.arange(
        self.config.n_expert, device=x.device)
    masks = masks.permute(2, 0, 1)  # (n_expert, B*T, n_expert_per_token)
    y = torch.zeros_like(x)  # (B*T, C)
    for mask, expert in zip(masks, self.experts):
      token_idx, expert_idx = torch.where(mask)
      y[token_idx] += probs[token_idx, expert_idx, None] * expert(x[token_idx])
    return y.view(B, T, C)


def build_rope_cache(
    seq_len: int,
    n_elem: int,
    device: Optional[torch.device] = None,
    base: int = 10000,
    condense_ratio: int = 1) -> Tuple[torch.Tensor, torch.Tensor]:
  """Enhanced Transformer with Rotary Position Embedding.

    Derived from: https://github.com/labmlai/annotated_deep_learning_paper_implementations/blob/master/labml_nn/
    transformers/rope/__init__.py. MIT License:
    https://github.com/labmlai/annotated_deep_learning_paper_implementations/blob/master/license.
    """
  # $\Theta = {\theta_i = 10000^{\frac{2(i-1)}{d}}, i \in [1, 2, ..., \frac{d}{2}]}$
  theta = 1.0 / (
      base**(torch.arange(0, n_elem, 2, device=device).float() / n_elem))

  # Create position indexes `[0, 1, ..., seq_len - 1]`
  seq_idx = torch.arange(seq_len, device=device) / condense_ratio

  # Calculate the product of position index and $\theta_i$
  idx_theta = torch.outer(seq_idx, theta).repeat(1, 2)

  return torch.cos(idx_theta), torch.sin(idx_theta)


def apply_rope(x: torch.Tensor, cos: torch.Tensor,
               sin: torch.Tensor) -> torch.Tensor:
  head_size = x.size(-1)
  x1 = x[..., :head_size // 2]  # (B, nh, T, hs/2)
  x2 = x[..., head_size // 2:]  # (B, nh, T, hs/2)
  rotated = torch.cat((-x2, x1), dim=-1)  # (B, nh, T, hs)
  roped = (x * cos) + (rotated * sin)
  return roped.to(dtype=x.dtype)


class KVCache(nn.Module):

  def __init__(
      self,
      k_shape: Tuple[int, int, int, int],
      v_shape: Tuple[int, int, int, int],
      device: Optional[torch.device] = None,
      dtype: Optional[torch.dtype] = None,
  ) -> None:
    super().__init__()
    self.register_buffer(
        "k", torch.zeros(k_shape, device=device, dtype=dtype), persistent=False)
    self.register_buffer(
        "v", torch.zeros(v_shape, device=device, dtype=dtype), persistent=False)

  def forward(self, input_pos: torch.Tensor, k: torch.Tensor,
              v: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
    # move the buffer to the activation dtype for when AMP is used
    self.k = self.k.to(k.dtype)
    self.v = self.v.to(v.dtype)
    # update the cache
    k = self.k.index_copy_(2, input_pos, k)
    v = self.v.index_copy_(2, input_pos, v)
    return k, v

  def reset_parameters(self) -> None:
    torch.nn.init.zeros_(self.k)
    torch.nn.init.zeros_(self.v)


def build_mask_cache(max_seq_length: int,
                     device: Optional[torch.device] = None) -> torch.Tensor:
  ones = torch.ones((max_seq_length, max_seq_length),
                    device=device,
                    dtype=torch.bool)
  return torch.tril(ones).unsqueeze(0).unsqueeze(0)


class RMSNorm(torch.nn.Module):
  """Root Mean Square Layer Normalization.

    Derived from https://github.com/bzhangGo/rmsnorm/blob/master/rmsnorm_torch.py. BSD 3-Clause License:
    https://github.com/bzhangGo/rmsnorm/blob/master/LICENSE.
    """

  def __init__(self,
               size: int,
               dim: int = -1,
               eps: float = 1e-6,
               add_unit_offset: bool = False) -> None:
    super().__init__()
    self.weight = torch.nn.Parameter(torch.ones(size))
    self.eps = eps
    self.dim = dim
    self.add_unit_offset = add_unit_offset

  def forward(self, x: torch.Tensor) -> torch.Tensor:
    dtype = x.dtype
    x = x.float()
    # NOTE: the original RMSNorm paper implementation is not equivalent
    norm_x = torch.mean(x * x, dim=self.dim, keepdim=True)
    x_normed = x * torch.rsqrt(norm_x + self.eps)
    x_normed = x_normed.to(dtype=dtype)
    if self.add_unit_offset:
      # Gemma model requires a unit offset
      # https://github.com/google/gemma_pytorch/blob/main/gemma/model.py#L176
      return x_normed * (1 + self.weight)
    return x_normed * self.weight

  def reset_parameters(self) -> None:
    torch.nn.init.ones_(self.weight)



# this version contains modification to make it easier to trace

import math
from dataclasses import dataclass
from typing import Any, Optional, Tuple, List

import torch
import torch.nn.functional as F
from torch import nn


@dataclass
class ModelArgs:
  dim: int = 4096
  n_layers: int = 32
  n_heads: int = 32
  n_kv_heads: Optional[int] = None
  vocab_size: int = -1  # defined later by tokenizer
  multiple_of: int = 256  # make SwiGLU hidden layer size multiple of large power of 2
  ffn_dim_multiplier: Optional[float] = None
  norm_eps: float = 1e-5
  max_batch_size: int = 32
  max_seq_len: int = 2048
  bf16_enable = True


class RMSNorm(torch.nn.Module):

  def __init__(self, dim: int, eps: float = 1e-6):
    super().__init__()
    self.eps = eps
    self.weight = nn.Parameter(torch.ones(dim))

  def _norm(self, x):
    return x * torch.rsqrt(x.pow(2).mean(-1, keepdim=True) + self.eps)

  def forward(self, x):
    output = self._norm(x.float()).type_as(x)
    return output * self.weight


def precompute_freqs_cis(dim: int, end: int, theta: float = 10000.0):
  freqs = 1.0 / (theta**(torch.arange(0, dim, 2)[:(dim // 2)].float() / dim))
  t = torch.arange(end, device=freqs.device)  # type: ignore
  freqs = torch.outer(t, freqs).float()  # type: ignore
  freqs_cis = torch.polar(torch.ones_like(freqs), freqs)  # complex64
  return freqs_cis


def reshape_for_broadcast(freqs_cis: torch.Tensor, x: torch.Tensor):
  ndim = x.ndim
  assert 0 <= 1 < ndim
  assert freqs_cis.shape == (x.shape[-3], x.shape[-1])
  shape = [d if i == 1 or i == ndim - 1 else 1 for i, d in enumerate(x.shape)]
  return freqs_cis.view(*shape)


def apply_rotary_emb(
    xq: torch.Tensor,
    xk: torch.Tensor,
    freqs_cis: torch.Tensor,
) -> Tuple[torch.Tensor, torch.Tensor]:
  # bs, seqlen, heads, dim
  xq_ = torch.view_as_complex(xq.float().reshape(*xq.shape[:-1], -1, 2))
  xk_ = torch.view_as_complex(xk.float().reshape(*xk.shape[:-1], -1, 2))
  freqs_cis = reshape_for_broadcast(freqs_cis, xq_)
  xq_out = torch.view_as_real(xq_ * freqs_cis).flatten(3)
  xk_out = torch.view_as_real(xk_ * freqs_cis).flatten(3)
  return xq_out.type_as(xq), xk_out.type_as(xk)


def repeat_kv(x: torch.Tensor, n_rep: int) -> torch.Tensor:
  """torch.repeat_interleave(x, dim=2, repeats=n_rep)"""
  bs, slen, n_kv_heads, head_dim = x.shape
  if n_rep == 1:
    return x
  return (x[:, :, :,
            None, :].expand(bs, slen, n_kv_heads, n_rep,
                            head_dim).reshape(bs, slen, n_kv_heads * n_rep,
                                              head_dim))


class Attention(nn.Module):

  def __init__(self, args: ModelArgs):
    super().__init__()

    self.n_kv_heads = args.n_heads if args.n_kv_heads is None else args.n_kv_heads
    self.n_local_heads = args.n_heads
    self.n_local_kv_heads = self.n_kv_heads
    self.n_rep = self.n_local_heads // self.n_local_kv_heads
    self.head_dim = args.dim // args.n_heads

    init_method = lambda x: x

    self.wq = nn.Linear(
        args.dim,
        args.n_heads * self.head_dim,
        bias=False,
    )
    self.wk = nn.Linear(
        args.dim,
        self.n_kv_heads * self.head_dim,
        bias=False,
    )
    self.wv = nn.Linear(
        args.dim,
        self.n_kv_heads * self.head_dim,
        bias=False,
    )
    self.wo = nn.Linear(
        args.n_heads * self.head_dim,
        args.dim,
        bias=False,
    )

  def forward(self, x: torch.Tensor, freqs_cis: torch.Tensor,
              mask: Optional[torch.Tensor], prefill: bool,
              input_indexes: torch.Tensor, cache_indexes: torch.Tensor, cache_k,
              cache_v):
    # bsz, seqlen, _ = x.shape
    bsz, seqlen = x.shape[0], x.shape[-2]
    xq, xk, xv = self.wq(x), self.wk(x), self.wv(x)

    xq = xq.view(bsz, seqlen, self.n_local_heads, self.head_dim)
    xk = xk.view(bsz, seqlen, self.n_local_kv_heads, self.head_dim)
    xv = xv.view(bsz, seqlen, self.n_local_kv_heads, self.head_dim)

    xq, xk = apply_rotary_emb(xq, xk, freqs_cis=freqs_cis)
    input_indexes = input_indexes.to(torch.int64)

    cache_k = cache_k.index_copy(1, input_indexes, xk)
    cache_v = cache_v.index_copy(1, input_indexes, xv)

    #keys = cache_k.index_select(1, cache_indexes)
    #values = cache_v.index_select(1, cache_indexes)
    keys = cache_k
    values = cache_v
    # repeat k/v heads if n_kv_heads < n_heads
    keys = repeat_kv(keys, self.n_rep)  # (bs, seqlen, n_local_heads, head_dim)
    values = repeat_kv(values,
                       self.n_rep)  # (bs, seqlen, n_local_heads, head_dim)

    #xq = xq.transpose(-3, -2)  # (bs, n_local_heads, seqlen, head_dim)
    #keys = keys.transpose(-3,-2)
    #values = values.transpose(-3,-2)
    xq_new = torch.clone(xq)
    #scores = torch.matmul(xq_new, keys.transpose(-3,-2).transpose(-2, -1)) / math.sqrt(self.head_dim)
    scores = torch.einsum('ijkl,imkl->ikjm', xq_new, keys) / math.sqrt(
        self.head_dim)
    if mask is not None:
      scores = scores + mask  # (bs, n_local_heads, seqlen, max_seqlen)
    scores = F.softmax(scores.float(), dim=-1).type_as(xq_new)
    #output = torch.matmul(scores, values.transpose(-3,-2))  # (bs, n_local_heads, seqlen, head_dim)
    output = torch.einsum('ikjm,imkl->ikjl', scores, values)
    output = output.transpose(-3, -2).contiguous().view(bsz, seqlen, -1)

    return self.wo(output), cache_k, cache_v


class FeedForward(nn.Module):

  def __init__(
      self,
      dim: int,
      hidden_dim: int,
      multiple_of: int,
      ffn_dim_multiplier: Optional[float],
  ):
    super().__init__()
    hidden_dim = int(2 * hidden_dim / 3)
    # custom dim factor multiplier
    if ffn_dim_multiplier is not None:
      hidden_dim = int(ffn_dim_multiplier * hidden_dim)
    hidden_dim = multiple_of * ((hidden_dim + multiple_of - 1) // multiple_of)

    init_method = lambda x: x

    self.w1 = nn.Linear(
        dim,
        hidden_dim,
        bias=False,
    )
    self.w2 = nn.Linear(
        hidden_dim,
        dim,
        bias=False,
    )
    self.w3 = nn.Linear(
        dim,
        hidden_dim,
        bias=False,
    )

  def forward(self, x):
    return self.w2(F.silu(self.w1(x)) * self.w3(x))


class TransformerBlock(nn.Module):

  def __init__(self,
               layer_id: int,
               args: ModelArgs,
               groups: Optional[List] = None):
    super().__init__()
    self.n_heads = args.n_heads
    self.dim = args.dim
    self.head_dim = args.dim // args.n_heads

    self.attention = Attention(args,)
    self.feed_forward = FeedForward(
        dim=args.dim,
        hidden_dim=4 * args.dim,
        multiple_of=args.multiple_of,
        ffn_dim_multiplier=args.ffn_dim_multiplier,
    )
    self.layer_id = layer_id
    self.attention_norm = RMSNorm(args.dim, eps=args.norm_eps)
    self.ffn_norm = RMSNorm(args.dim, eps=args.norm_eps)

  def forward(
      self,
      x: torch.Tensor,
      freqs_cis: torch.Tensor,
      mask: Optional[torch.Tensor],
      prefill: bool,
      input_indexes: torch.Tensor,
      cache_indexes,
      cache_k,
      cache_v,
  ):
    attn, xk, xv = self.attention.forward(
        self.attention_norm(x), freqs_cis, mask, prefill, input_indexes,
        cache_indexes, cache_k, cache_v)
    h = x + attn
    out = h + self.feed_forward.forward(self.ffn_norm(h))
    return out, xk, xv


class Transformer(nn.Module):

  def __init__(self,
               params: ModelArgs,
               world_size: Optional[int] = None,
               rank: Optional[int] = None,
               groups: Optional[List] = None):
    super().__init__()
    self.params = params
    self.vocab_size = params.vocab_size
    self.n_layers = params.n_layers

    init_method = lambda x: x

    self.tok_embeddings = nn.Embedding(
        params.vocab_size,
        params.dim,
    )

    self.layers = torch.nn.ModuleList()
    for layer_id in range(params.n_layers):
      self.layers.append(TransformerBlock(
          layer_id,
          params,
      ))

    self.norm = RMSNorm(params.dim, eps=params.norm_eps)
    self.output = nn.Linear(
        params.dim,
        params.vocab_size,
        bias=False,
    )

    freqs_cis = precompute_freqs_cis(self.params.dim // self.params.n_heads,
                                     self.params.max_seq_len * 2)
    # self.register_buffer("freqs_cis", freqs_cis)
    self.freqs_cis = freqs_cis
    mask = torch.full(
        (1, 1, self.params.max_seq_len, self.params.max_seq_len),
        float("-inf")).to(
            torch.bfloat16 if self.params.bf16_enable else torch.float)
    mask = torch.triu(mask, diagonal=1)
    self.mask = mask
    # self.register_buffer("mask", mask)

  @torch.no_grad()
  def forward(self, tokens: torch.Tensor, input_indexes: torch.Tensor,
              cache_indexes, caches: List[Tuple[torch.tensor, ...]], prefill):
    seqlen = tokens.shape[-1]
    h = self.tok_embeddings(tokens)
    freqs_cis = self.freqs_cis.index_select(0, input_indexes)
    mask = None
    if prefill:
      mask = torch.full((1, 1, seqlen, seqlen), float("-inf")).to(
          torch.bfloat16 if self.params.bf16_enable else torch.float)
      mask = torch.triu(mask, diagonal=1)

    new_caches = []
    for layer, (cache_k, cache_v) in zip(self.layers, caches):
      h, new_k, new_v = layer(h, freqs_cis, mask, prefill, input_indexes,
                              cache_indexes, cache_k, cache_v)
      new_caches.append((new_k, new_v))
    h = self.norm(h)
    output = self.output(h).float()
    return output, new_caches



from absl.testing import parameterized
import unittest
from typing import Tuple
import itertools
from functools import partial
import jax
import torch

import torchax
import torchax.interop


@partial(jax.jit, static_argnums=(1, 2, 3, 4))
def upsample_jit(tensor, output_size: Tuple[int, int], align_corners: bool,
                 antialias: bool, method: str):
  tensor = torchax.interop.torch_view(tensor)
  tensor = torch.nn.functional.interpolate(
      tensor,
      size=output_size,
      mode=method,
      align_corners=align_corners,
      antialias=antialias)
  return torchax.interop.jax_view(tensor)


class TestResampling(parameterized.TestCase):

  @parameterized.product(
      antialias=[
          True,
          False,
      ], align_corners=[
          False,
          True,
      ])
  def test_resampling_combinations_bicubic(self, antialias, align_corners):
    method = 'bicubic'
    input_tensor = torch.rand((1, 1, 256, 512), dtype=torch.float32)
    output_size = (128, 64)

    upsampled_tensor = torch.nn.functional.interpolate(
        input_tensor,
        size=output_size,
        mode=method,
        align_corners=align_corners,
        antialias=antialias)

    env = torchax.default_env()
    with env:
      input_tensor_xla = env.to_xla(input_tensor)
      input_tensor_xla = torchax.interop.jax_view(input_tensor_xla)
      upsampled_tensor_xla = upsample_jit(
          input_tensor_xla,
          output_size,
          align_corners,
          antialias=antialias,
          method=method)

    upsampled_tensor_xla = env.j2t_copy(upsampled_tensor_xla)
    abs_err = torch.abs(upsampled_tensor - upsampled_tensor_xla)

    assert torch.allclose(
        upsampled_tensor, upsampled_tensor_xla, atol=1e-4,
        rtol=1e-5), f"{method} upsampling failed with error {abs_err.max()}"


if __name__ == '__main__':
  unittest.main()
